# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

#
# These tests are adapted from ntcore's test suite
#

import pytest

from threading import Condition

from _pynetworktables._impl.constants import NT_NOTIFY_LOCAL, NT_NOTIFY_NEW

from _pynetworktables._impl.value import Value


class SC(object):
    def __init__(self):
        self.events = []
        self.event_cond = Condition()

    def __call__(self, event):
        with self.event_cond:
            self.events.append(event)
            self.event_cond.notify()

    def wait(self, count):
        with self.event_cond:
            result = self.event_cond.wait_for(lambda: len(self.events) == count, 2)
            assert result, "expected %s events, got %s" % (count, len(self.events))
            return self.events[:]


@pytest.fixture
def server_cb():
    return SC()


def test_EntryNewLocal(nt_live, server_cb):
    nt_server, nt_client = nt_live
    nt_server_api = nt_server._api

    nt_server_api.addEntryListenerById(
        nt_server_api.getEntryId("/foo"), server_cb, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL
    )

    # Trigger an event
    nt_server_api.setEntryValueById(
        nt_server_api.getEntryId("/foo/bar"), Value.makeDouble(2.0)
    )
    nt_server_api.setEntryValueById(
        nt_server_api.getEntryId("/foo"), Value.makeDouble(1.0)
    )

    assert nt_server_api.waitForEntryListenerQueue(1.0)

    # Check the event
    events = server_cb.wait(1)

    # assert events[0].listener == handle
    assert events[0].local_id == nt_server_api.getEntryId("/foo")
    assert events[0].name == "/foo"
    assert events[0].value == Value.makeDouble(1.0)
    assert events[0].flags == NT_NOTIFY_NEW | NT_NOTIFY_LOCAL


def test_EntryNewRemote(nt_live, server_cb):
    nt_server, nt_client = nt_live
    nt_server_api = nt_server._api
    nt_client_api = nt_client._api

    nt_server_api.addEntryListenerById(
        nt_server_api.getEntryId("/foo"), server_cb, NT_NOTIFY_NEW
    )

    # Trigger an event
    nt_client_api.setEntryValueById(
        nt_client_api.getEntryId("/foo/bar"), Value.makeDouble(2.0)
    )
    nt_client_api.setEntryValueById(
        nt_client_api.getEntryId("/foo"), Value.makeDouble(1.0)
    )

    nt_client_api.flush()

    assert nt_server_api.waitForEntryListenerQueue(1.0)

    # Check the event
    events = server_cb.wait(1)

    # assert events[0].listener == handle
    assert events[0].local_id == nt_server_api.getEntryId("/foo")
    assert events[0].name == "/foo"
    assert events[0].value == Value.makeDouble(1.0)
    assert events[0].flags == NT_NOTIFY_NEW


def test_PrefixNewLocal(nt_live, server_cb):
    nt_server, nt_client = nt_live
    nt_server_api = nt_server._api

    nt_server_api.addEntryListener("/foo", server_cb, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)

    # Trigger an event
    nt_server_api.setEntryValueById(
        nt_server_api.getEntryId("/foo/bar"), Value.makeDouble(1.0)
    )
    nt_server_api.setEntryValueById(
        nt_server_api.getEntryId("/baz"), Value.makeDouble(1.0)
    )

    assert nt_server_api.waitForEntryListenerQueue(1.0)

    events = server_cb.wait(1)

    # assert events[0].listener == handle
    assert events[0].local_id == nt_server_api.getEntryId("/foo/bar")
    assert events[0].name == "/foo/bar"
    assert events[0].value == Value.makeDouble(1.0)
    assert events[0].flags == NT_NOTIFY_NEW | NT_NOTIFY_LOCAL


def test_PrefixNewRemote(nt_live, server_cb):
    nt_server, nt_client = nt_live
    nt_server_api = nt_server._api
    nt_client_api = nt_client._api

    nt_server_api.addEntryListener("/foo", server_cb, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)

    # Trigger an event
    nt_client_api.setEntryValueById(
        nt_client_api.getEntryId("/foo/bar"), Value.makeDouble(1.0)
    )
    nt_client_api.setEntryValueById(
        nt_client_api.getEntryId("/baz"), Value.makeDouble(1.0)
    )

    assert nt_server_api.waitForEntryListenerQueue(1.0)

    # Check the event
    events = server_cb.wait(1)

    # assert events[0].listener == handle
    assert events[0].local_id == nt_server_api.getEntryId("/foo/bar")
    assert events[0].name == "/foo/bar"
    assert events[0].value == Value.makeDouble(1.0)
    assert events[0].flags == NT_NOTIFY_NEW



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os

from os.path import abspath, join, dirname

sys.path.insert(0, abspath(dirname(__file__)))
sys.path.insert(0, abspath(join(dirname(__file__), "..")))

import networktables

# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This is used for linking and such so we link to the thing we're building
rtd_version = os.environ.get("READTHEDOCS_VERSION", "latest")
if rtd_version not in ["stable", "latest"]:
    rtd_version = "stable"

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]

intersphinx_mapping = {
    "robotpy": ("http://robotpy.readthedocs.io/en/%s/" % rtd_version, None),
    "wpilib": ("http://robotpy-wpilib.readthedocs.io/en/%s/" % rtd_version, None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "RobotPy NetworkTables"
copyright = "2016, RobotPy Development Team"


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(networktables.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = networktables.__version__

autoclass_content = "both"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "default"

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
else:
    html_theme = "default"

# Output file base name for HTML help builder.
htmlhelp_basename = "networktablesdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        "networktables.tex",
        "RobotPy networktables Documentation",
        "RobotPy Development Team",
        "manual",
    )
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "index",
        "networktables",
        "RobotPy networktables Documentation",
        ["RobotPy Development Team"],
        1,
    )
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "networktables",
        "RobotPy networktables Documentation",
        "RobotPy Development Team",
        "networktables",
        "One line description of project.",
        "Miscellaneous",
    )
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "RobotPy NetworkTables"
epub_author = "RobotPy Development Team"
epub_publisher = "RobotPy Development Team"
epub_copyright = "2014, RobotPy Development Team"

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Custom Document processing ----------------------------------------------

import gensidebar

gensidebar.generate_sidebar(globals(), "pynetworktables")



#!/usr/bin/env python3
#
# Sample RobotPy program that demonstrates how to send data to the logger
# example program
#
# This is NOT REQUIRED to use the example, any robot language can be used!
# Translate the necessary bits into your own robot code.s
#

import wpilib


class MyRobot(wpilib.IterativeRobot):
    """Main robot class"""

    def robotInit(self):
        """Robot-wide initialization code should go here"""

    def autonomousInit(self):
        """Called only at the beginning of autonomous mode"""
        self.useless = 1

    def autonomousPeriodic(self):
        """Called every 20ms in autonomous mode"""
        self.useless += 1

        # Obviously, this is fabricated... do something more useful!
        data1 = self.useless
        data2 = self.useless * 2

        # Only write once per loop
        wpilib.SmartDashboard.putNumberArray(
            "log_data", [wpilib.Timer.getFPGATimestamp(), data1, data2]
        )

    def disabledInit(self):
        """Called only at the beginning of disabled mode"""
        pass

    def disabledPeriodic(self):
        """Called every 20ms in disabled mode"""
        pass

    def teleopInit(self):
        """Called only at the beginning of teleoperated mode"""
        pass

    def teleopPeriodic(self):
        """Called every 20ms in teleoperated mode"""


if __name__ == "__main__":
    wpilib.run(MyRobot)



# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

#
# These tests are adapted from ntcore's test suite
#

from io import StringIO
from unittest.mock import call, Mock, ANY

import pytest

from _pynetworktables._impl.constants import (
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    NT_PERSISTENT,
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS,
)

from _pynetworktables._impl.dispatcher import Dispatcher
from _pynetworktables._impl.entry_notifier import EntryNotifier
from _pynetworktables._impl.message import Message
from _pynetworktables._impl.network_connection import NetworkConnection
from _pynetworktables._impl.rpc_server import RpcServer
from _pynetworktables._impl.storage import Storage
from _pynetworktables._impl.value import Value


@pytest.fixture
def dispatcher():
    return Mock(spec=Dispatcher)


@pytest.fixture
def entry_notifier():
    return Mock(spec=EntryNotifier)


@pytest.fixture(params=[True, False])
def is_server(request):
    return request.param


@pytest.fixture
def conn():
    conn = Mock(spec=NetworkConnection)
    conn.get_proto_rev.return_value = 0x0300
    return conn


class FakeUserEntry:
    def __init__(self, *args):
        pass


@pytest.fixture
def storage_empty(dispatcher, entry_notifier, is_server):
    rpc_server = Mock(spec=RpcServer)

    storage = Storage(entry_notifier, rpc_server, FakeUserEntry)
    storage.setDispatcher(dispatcher, is_server)

    entry_notifier.m_local_notifiers = True

    yield storage

    rpc_server.stop()
    storage.stop()


@pytest.fixture
def storage_populate_one(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    entry_notifier.m_local_notifiers = False

    storage.setEntryTypeValue("foo", Value.makeBoolean(True))

    dispatcher.reset_mock()
    entry_notifier.reset_mock()
    entry_notifier.m_local_notifiers = True

    return storage


@pytest.fixture
def storage_populated(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    entry_notifier.m_local_notifiers = False

    entry_notifier.m_local_notifiers = False
    storage.setEntryTypeValue("foo", Value.makeBoolean(True))
    storage.setEntryTypeValue("foo2", Value.makeDouble(0.0))
    storage.setEntryTypeValue("bar", Value.makeDouble(1.0))
    storage.setEntryTypeValue("bar2", Value.makeBoolean(False))

    dispatcher.reset_mock()
    entry_notifier.reset_mock()
    entry_notifier.m_local_notifiers = True

    return storage


@pytest.fixture
def storage_persistent(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    entry_notifier.m_local_notifiers = False

    storage.setEntryTypeValue("boolean/true", Value.makeBoolean(True))
    storage.setEntryTypeValue("boolean/false", Value.makeBoolean(False))
    storage.setEntryTypeValue("double/neg", Value.makeDouble(-1.5))
    storage.setEntryTypeValue("double/zero", Value.makeDouble(0.0))
    storage.setEntryTypeValue("double/big", Value.makeDouble(1.3e8))
    storage.setEntryTypeValue("string/empty", Value.makeString(""))
    storage.setEntryTypeValue("string/normal", Value.makeString("hello"))
    storage.setEntryTypeValue("string/special", Value.makeString("\0\3\5\n"))
    storage.setEntryTypeValue("string/quoted", Value.makeString('"a"'))
    storage.setEntryTypeValue("raw/empty", Value.makeRaw(b""))
    storage.setEntryTypeValue("raw/normal", Value.makeRaw(b"hello"))
    storage.setEntryTypeValue("raw/special", Value.makeRaw(b"\0\3\5\n"))
    storage.setEntryTypeValue("booleanarr/empty", Value.makeBooleanArray([]))
    storage.setEntryTypeValue("booleanarr/one", Value.makeBooleanArray([True]))
    storage.setEntryTypeValue("booleanarr/two", Value.makeBooleanArray([True, False]))
    storage.setEntryTypeValue("doublearr/empty", Value.makeDoubleArray([]))
    storage.setEntryTypeValue("doublearr/one", Value.makeDoubleArray([0.5]))
    storage.setEntryTypeValue("doublearr/two", Value.makeDoubleArray([0.5, -0.25]))
    storage.setEntryTypeValue("stringarr/empty", Value.makeStringArray([]))
    storage.setEntryTypeValue("stringarr/one", Value.makeStringArray(["hello"]))
    storage.setEntryTypeValue(
        "stringarr/two", Value.makeStringArray(["hello", "world\n"])
    )
    storage.setEntryTypeValue("\0\3\5\n", Value.makeBoolean(True))
    storage.setEntryTypeValue("CaseSensitive/KeyName", Value.makeBoolean(True))
    storage.setEntryTypeValue("=", Value.makeBoolean(True))

    dispatcher.reset_mock()
    entry_notifier.reset_mock()
    entry_notifier.m_local_notifiers = True

    return storage


def test_Construct(storage_empty):
    storage = storage_empty

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0


def test_StorageEntryInit(storage_empty):
    storage = storage_empty

    assert storage.m_entries.get("foo") is None


def test_GetEntryValueNotExist(storage_empty, dispatcher):
    storage = storage_empty

    assert storage.getEntryValue("foo") is None
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0
    assert dispatcher._queueOutgoing.call_count == 0


def test_GetEntryValueExist(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    value = Value.makeBoolean(True)
    storage.setEntryTypeValue("foo", value)
    assert dispatcher._queueOutgoing.call_count == 1
    assert entry_notifier.notifyEntry.call_count == 1

    assert value == storage.getEntryValue("foo")


def test_SetEntryTypeValueAssignNew(
    storage_empty, dispatcher, entry_notifier, is_server
):
    storage = storage_empty

    # brand new entry
    value = Value.makeBoolean(True)
    storage.setEntryTypeValue("foo", value)
    assert value == storage.m_entries.get("foo").value

    dispatcher._queueOutgoing.assert_has_calls(
        [
            call(
                Message.entryAssign("foo", 0 if is_server else 0xFFFF, 1, value, 0),
                None,
                None,
            )
        ]
    )
    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)]
    )

    if is_server:
        assert 1 == len(storage.m_idmap)
        assert value == storage.m_idmap[0].value
    else:
        assert len(storage.m_idmap) == 0


def test_SetEntryTypeValueAssignTypeChange(
    storage_populate_one, dispatcher, entry_notifier, is_server
):
    storage = storage_populate_one

    # update with different type results in assignment message
    value = Value.makeDouble(0.0)
    storage.setEntryTypeValue("foo", value)
    assert value == storage.m_entries.get("foo").value

    dispatcher._queueOutgoing.assert_has_calls(
        [
            call(
                Message.entryAssign("foo", 0 if is_server else 0xFFFF, 2, value, 0),
                None,
                None,
            )
        ]
    )
    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)]
    )


def test_SetEntryTypeValueEqualValue(storage_populate_one, dispatcher, entry_notifier):
    storage = storage_populate_one

    # update with same type and same value: change value contents but no update
    # message is issued (minimizing bandwidth usage)
    value = Value.makeBoolean(True)
    storage.setEntryTypeValue("foo", value)
    assert value == storage.m_entries.get("foo").value

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryTypeValueDifferentValue(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    # update with same type and different value results in value update message
    value = Value.makeDouble(1.0)
    storage.setEntryTypeValue("foo2", value)
    assert value == storage.m_entries.get("foo2").value

    if is_server:
        # id assigned if server; seq_num incremented
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryUpdate(1, 2, value), None, None)]
        )

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)]
    )

    if not is_server:
        # shouldn't send an update id not assigned yet (happens on client only)
        assert dispatcher._queueOutgoing.call_count == 0
        assert 2 == storage.m_entries.get("foo2").seq_num  # still should be incremented


def test_SetEntryTypeValueEmptyName(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    value = Value.makeBoolean(True)
    storage.setEntryTypeValue("", value)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryTypeValueEmptyValue(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    storage.setEntryTypeValue("foo", None)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryValueAssignNew(storage_empty, dispatcher, entry_notifier, is_server):
    storage = storage_empty

    # brand entry
    value = Value.makeBoolean(True)
    assert storage.setEntryValue("foo", value)
    assert value == storage.m_entries.get("foo").value

    dispatcher._queueOutgoing.assert_has_calls(
        [
            call(
                Message.entryAssign("foo", 0 if is_server else 0xFFFF, 1, value, 0),
                None,
                None,
            )
        ]
    )
    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)]
    )


def test_SetEntryValueAssignTypeChange(
    storage_populate_one, dispatcher, entry_notifier
):
    storage = storage_populate_one

    # update with different type results in error and no message
    value = Value.makeDouble(0.0)
    assert not storage.setEntryValue("foo", value)
    entry = storage.m_entries.get("foo")
    assert value != entry.value

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryValueEqualValue(storage_populate_one, dispatcher, entry_notifier):
    storage = storage_populate_one

    # update with same type and same value: change value contents but no update
    # message is issued (minimizing bandwidth usage)
    value = Value.makeBoolean(True)
    assert storage.setEntryValue("foo", value)
    entry = storage.m_entries.get("foo")
    assert value == entry.value

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryValueDifferentValue(
    storage_populated, is_server, dispatcher, entry_notifier
):
    storage = storage_populated

    # update with same type and different value results in value update message
    value = Value.makeDouble(1.0)
    assert storage.setEntryValue("foo2", value)
    entry = storage.m_entries.get("foo2")
    assert value == entry.value

    # client shouldn't send an update as id not assigned yet
    if is_server:
        # id assigned if server; seq_num incremented
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryUpdate(1, 2, value), None, None)]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)]
    )

    if not is_server:
        assert 2 == storage.m_entries.get("foo2").seq_num  # still should be incremented


def test_SetEntryValueEmptyName(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    value = Value.makeBoolean(True)
    assert storage.setEntryValue("", value)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryValueEmptyValue(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    assert storage.setEntryValue("foo", None)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetDefaultEntryAssignNew(storage_empty, dispatcher, entry_notifier, is_server):
    storage = storage_empty

    # brand new entry
    value = Value.makeBoolean(True)
    ret_val = storage.setDefaultEntryValue("foo", value)
    assert ret_val
    assert value == storage.m_entries.get("foo").value

    dispatcher._queueOutgoing.assert_has_calls(
        [
            call(
                Message.entryAssign("foo", 0 if is_server else 0xFFFF, 1, value, 0),
                None,
                None,
            )
        ]
    )
    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)]
    )


def test_SetDefaultEntryExistsSameType(
    storage_populate_one, dispatcher, entry_notifier
):
    storage = storage_populate_one

    # existing entry
    value = Value.makeBoolean(False)
    ret_val = storage.setDefaultEntryValue("foo", value)
    assert ret_val
    assert value != storage.m_entries.get("foo").value

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetDefaultEntryExistsDifferentType(
    storage_populate_one, dispatcher, entry_notifier
):
    storage = storage_populate_one

    # existing entry is boolean
    value = Value.makeDouble(2.0)
    ret_val = storage.setDefaultEntryValue("foo", value)
    assert not ret_val
    # should not have updated value in table if it already existed.
    assert value != storage.m_entries.get("foo").value

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_empty_SetDefaultEntryEmptyName(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    value = Value.makeBoolean(True)
    ret_val = storage.setDefaultEntryValue("", value)
    assert not ret_val
    assert "foo" not in storage.m_entries

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_empty_SetDefaultEntryEmptyValue(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    # value = Value.makeBoolean(True)
    ret_val = storage.setDefaultEntryValue("", None)
    assert not ret_val
    assert "foo" not in storage.m_entries

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_populated_SetDefaultEntryEmptyName(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    value = Value.makeBoolean(True)
    ret_val = storage.setDefaultEntryValue("", value)
    assert not ret_val
    # assert that no entries get added
    assert 4 == len(storage.m_entries)
    if is_server:
        assert 4 == len(storage.m_idmap)

    else:
        assert 0 == len(storage.m_idmap)

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_populated_SetDefaultEntryEmptyValue(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    # value = Value.makeBoolean(True)
    ret_val = storage.setDefaultEntryValue("", None)
    assert not ret_val
    # assert that no entries get added
    assert 4 == len(storage.m_entries)
    if is_server:
        assert 4 == len(storage.m_idmap)

    else:
        assert 0 == len(storage.m_idmap)

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryFlagsNew(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    # flags setting doesn't create an entry
    storage.setEntryFlags("foo", 0)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryFlagsEqualValue(storage_populate_one, dispatcher, entry_notifier):
    storage = storage_populate_one

    # update with same value: no update message is issued (minimizing bandwidth
    # usage)
    storage.setEntryFlags("foo", 0)
    entry = storage.m_entries.get("foo")
    assert 0 == entry.flags

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_SetEntryFlagsDifferentValue(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    # update with different value results in flags update message
    storage.setEntryFlags("foo2", 1)
    entry = storage.m_entries.get("foo2")
    assert 1 == entry.flags

    if is_server:
        # id assigned as this is the server
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.flagsUpdate(1, 1), None, None)]
        )
    else:
        # shouldn't send an update id not assigned yet
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", entry.value, NT_NOTIFY_FLAGS | NT_NOTIFY_LOCAL)]
    )


def test_SetEntryFlagsEmptyName(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    storage.setEntryFlags("", 0)
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_GetEntryFlagsNotExist(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    assert 0 == storage.getEntryFlags("foo")
    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0
    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_GetEntryFlagsExist(
    storage_populate_one, dispatcher, entry_notifier, is_server
):
    storage = storage_populate_one

    storage.setEntryFlags("foo", 1)
    assert 1 == storage.getEntryFlags("foo")

    assert dispatcher._queueOutgoing.call_count == (1 if is_server else 0)
    assert entry_notifier.notifyEntry.call_count == 1


def test_DeleteEntryNotExist(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    storage.deleteEntry("foo")

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_DeleteEntryExist(storage_populated, dispatcher, entry_notifier, is_server):
    storage = storage_populated

    storage.deleteEntry("foo2")

    entry = storage.m_entries.get("foo2")
    assert entry is not None
    assert entry.value is None
    assert entry.id == 0xFFFF
    assert entry.local_write == False

    # client shouldn't send an update as id not assigned yet
    if is_server:
        # id assigned as this is the server
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryDelete(1), None, None)]
        )
    else:
        # shouldn't send an update id not assigned yet
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", Value.makeDouble(0.0), NT_NOTIFY_DELETE | NT_NOTIFY_LOCAL)]
    )

    if is_server:
        assert len(storage.m_idmap) >= 2
        assert not storage.m_idmap[1]


def test_DeleteAllEntriesEmpty(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    storage.deleteAllEntries()
    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_deleteAllEntries(storage_populated, dispatcher, entry_notifier):
    storage = storage_populated

    storage.deleteAllEntries()
    assert len(storage.m_entries) == 4
    assert len(storage.getEntries("", 0)) == 0

    assert storage.m_entries["foo2"].value is None

    dispatcher._queueOutgoing.assert_has_calls(
        [call(Message.clearEntries(), None, None)]
    )

    entry_notifier.notifyEntry.assert_has_calls(
        [call(ANY, ANY, ANY, NT_NOTIFY_DELETE | NT_NOTIFY_LOCAL)] * 4
    )


def test_DeleteAllEntriesPersistent(storage_populated, dispatcher, entry_notifier):
    storage = storage_populated

    entry = storage.m_entries.get("foo2")
    entry.flags = NT_PERSISTENT
    entry.isPersistent = True
    storage.deleteAllEntries()
    assert len(storage.getEntries("", 0)) == 1
    assert "foo2" in storage.m_entries

    dispatcher._queueOutgoing.assert_has_calls(
        [call(Message.clearEntries(), None, None)]
    )

    entry_notifier.notifyEntry.assert_has_calls(
        [call(ANY, ANY, ANY, NT_NOTIFY_DELETE | NT_NOTIFY_LOCAL)] * 3
    )


def test_GetEntryInfoAll(storage_populated):
    storage = storage_populated

    info = storage.getEntryInfo("", 0)
    assert 4 == len(info)


def test_GetEntryInfoPrefix(storage_populated):
    storage = storage_populated

    info = storage.getEntryInfo("foo", 0)
    assert 2 == len(info)
    if info[0].name == "foo":
        assert "foo" == info[0].name
        assert NT_BOOLEAN == info[0].type
        assert "foo2" == info[1].name
        assert NT_DOUBLE == info[1].type

    else:
        assert "foo2" == info[0].name
        assert NT_DOUBLE == info[0].type
        assert "foo" == info[1].name
        assert NT_BOOLEAN == info[1].type


def test_GetEntryInfoTypes(storage_populated):
    storage = storage_populated

    info = storage.getEntryInfo("", NT_DOUBLE)
    assert 2 == len(info)
    assert NT_DOUBLE == info[0].type
    assert NT_DOUBLE == info[1].type
    if info[0].name == "foo2":
        assert "foo2" == info[0].name
        assert "bar" == info[1].name

    else:
        assert "bar" == info[0].name
        assert "foo2" == info[1].name


def test_GetEntryInfoPrefixTypes(storage_populated):
    storage = storage_populated

    info = storage.getEntryInfo("bar", NT_BOOLEAN)
    assert 1 == len(info)
    assert "bar2" == info[0].name
    assert NT_BOOLEAN == info[0].type


def test_SavePersistentEmpty(storage_persistent):
    storage = storage_persistent

    fp = StringIO()
    storage.savePersistent(fp=fp, periodic=False)

    fp.seek(0)
    assert "[NetworkTables Storage 3.0]\n\n" == fp.read()


def test_savePersistent(storage_persistent):
    storage = storage_persistent

    for e in storage.m_entries.values():
        e.flags = NT_PERSISTENT
        e.isPersistent = True

    fp = StringIO()
    storage.savePersistent(fp=fp, periodic=False)

    fp.seek(0)

    line = fp.readline()[:-1]
    assert "[NetworkTables Storage 3.0]" == line
    line = fp.readline()[:-1]
    assert 'boolean "\\x00\\x03\\x05\\n"=true' == line
    line = fp.readline()[:-1]
    assert 'boolean "="=true' == line
    line = fp.readline()[:-1]
    assert 'boolean "CaseSensitive/KeyName"=true' == line
    line = fp.readline()[:-1]
    assert 'boolean "boolean/false"=false' == line
    line = fp.readline()[:-1]
    assert 'boolean "boolean/true"=true' == line
    line = fp.readline()[:-1]
    assert 'array boolean "booleanarr/empty"=' == line
    line = fp.readline()[:-1]
    assert 'array boolean "booleanarr/one"=true' == line
    line = fp.readline()[:-1]
    assert 'array boolean "booleanarr/two"=true,false' == line
    line = fp.readline()[:-1]
    # this differs from ntcore
    assert 'double "double/big"=130000000.0' == line
    line = fp.readline()[:-1]
    assert 'double "double/neg"=-1.5' == line
    line = fp.readline()[:-1]
    assert 'double "double/zero"=0.0' == line
    line = fp.readline()[:-1]
    assert 'array double "doublearr/empty"=' == line
    line = fp.readline()[:-1]
    assert 'array double "doublearr/one"=0.5' == line
    line = fp.readline()[:-1]
    assert 'array double "doublearr/two"=0.5,-0.25' == line
    line = fp.readline()[:-1]
    assert 'raw "raw/empty"=' == line
    line = fp.readline()[:-1]
    assert 'raw "raw/normal"=aGVsbG8=' == line
    line = fp.readline()[:-1]
    assert 'raw "raw/special"=AAMFCg==' == line
    line = fp.readline()[:-1]
    assert 'string "string/empty"=""' == line
    line = fp.readline()[:-1]
    assert 'string "string/normal"="hello"' == line
    line = fp.readline()[:-1]
    assert 'string "string/quoted"="\\"a\\""' == line
    line = fp.readline()[:-1]
    assert 'string "string/special"="\\x00\\x03\\x05\\n"' == line
    line = fp.readline()[:-1]
    assert 'array string "stringarr/empty"=' == line
    line = fp.readline()[:-1]
    assert 'array string "stringarr/one"="hello"' == line
    line = fp.readline()[:-1]
    assert 'array string "stringarr/two"="hello","world\\n"' == line
    line = fp.readline()[:-1]
    assert "" == line


def test_LoadPersistentBadHeader(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    # EXPECT_CALL(warn, Warn(1, llvm.StringRef("header line mismatch, rest of file")))
    fp = StringIO(None)
    assert storage.loadPersistent(fp=fp) is not None

    fp = StringIO("[NetworkTables")
    # EXPECT_CALL(warn, Warn(1, llvm.StringRef("header line mismatch, rest of file")))
    assert storage.loadPersistent(fp=fp) is not None

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0
    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_LoadPersistentCommentHeader(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    fp = StringIO("\n; comment\n# comment\n[NetworkTables Storage 3.0]\n")
    assert storage.loadPersistent(fp=fp) is None

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0
    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_LoadPersistentEmptyName(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    fp = StringIO('[NetworkTables Storage 3.0]\nboolean ""=true\n')
    assert storage.loadPersistent(fp=fp) is None

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0
    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_LoadPersistentAssign(storage_empty, dispatcher, entry_notifier, is_server):
    storage = storage_empty

    fp = StringIO('[NetworkTables Storage 3.0]\nboolean "foo"=true\n')
    assert storage.loadPersistent(fp=fp) is None

    entry = storage.m_entries.get("foo")
    assert Value.makeBoolean(True) == entry.value
    assert NT_PERSISTENT == entry.flags
    assert entry.isPersistent

    dispatcher._queueOutgoing.assert_has_calls(
        [
            call(
                Message.entryAssign(
                    "foo", 0 if is_server else 0xFFFF, 1, entry.value, NT_PERSISTENT
                ),
                None,
                None,
            )
        ]
    )

    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", entry.value, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)]
    )


def test_LoadPersistentUpdateFlags(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    fp = StringIO('[NetworkTables Storage 3.0]\ndouble "foo2"=0.0\n')
    assert storage.loadPersistent(fp=fp) is None

    entry = storage.m_entries.get("foo2")
    assert Value.makeDouble(0.0) == entry.value
    assert NT_PERSISTENT == entry.flags
    assert entry.isPersistent

    if is_server:
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.flagsUpdate(1, NT_PERSISTENT), None, None)]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", entry.value, NT_NOTIFY_FLAGS | NT_NOTIFY_LOCAL)]
    )


def test_LoadPersistentUpdateValue(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    entry = storage.m_entries.get("foo2")
    entry.flags = NT_PERSISTENT
    entry.isPersistent = True

    fp = StringIO('[NetworkTables Storage 3.0]\ndouble "foo2"=1.0\n')
    assert storage.loadPersistent(fp=fp) is None

    entry = storage.m_entries.get("foo2")
    assert Value.makeDouble(1.0) == entry.value
    assert NT_PERSISTENT == entry.flags
    assert entry.isPersistent

    # client shouldn't send an update as id not assigned yet
    if is_server:
        # id assigned as this is the server; seq_num incremented
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryUpdate(1, 2, entry.value), None, None)]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(1, "foo2", entry.value, NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL)]
    )

    if not is_server:
        assert 2 == storage.m_entries.get("foo2").seq_num  # still should be incremented


def test_LoadPersistentUpdateValueFlags(
    storage_populated, dispatcher, entry_notifier, is_server
):
    storage = storage_populated

    fp = StringIO('[NetworkTables Storage 3.0]\ndouble "foo2"=1.0\n')
    assert storage.loadPersistent(fp=fp) is None

    entry = storage.m_entries.get("foo2")
    assert Value.makeDouble(1.0) == entry.value
    assert NT_PERSISTENT == entry.flags
    assert entry.isPersistent

    # client shouldn't send an update as id not assigned yet
    if is_server:
        # id assigned as this is the server; seq_num incremented
        dispatcher._queueOutgoing.assert_has_calls(
            [
                call(Message.entryUpdate(1, 2, entry.value), None, None),
                call(Message.flagsUpdate(1, NT_PERSISTENT), None, None),
            ]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [
            call(
                1,
                "foo2",
                entry.value,
                NT_NOTIFY_FLAGS | NT_NOTIFY_UPDATE | NT_NOTIFY_LOCAL,
            )
        ]
    )

    if not is_server:
        assert 2 == storage.m_entries.get("foo2").seq_num  # still should be incremented


def test_loadPersistent(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    inp = "[NetworkTables Storage 3.0]\n"
    inp += 'boolean "\\x00\\x03\\x05\\n"=true\n'
    inp += 'boolean "CaseSensitive/KeyName"=true\n'
    inp += 'boolean "boolean/false"=false\n'
    inp += 'boolean "boolean/true"=true\n'
    inp += 'array boolean "booleanarr/empty"=\n'
    inp += 'array boolean "booleanarr/one"=true\n'
    inp += 'array boolean "booleanarr/two"=true,false\n'
    inp += 'double "double/big"=1.3e+08\n'
    inp += 'double "double/neg"=-1.5\n'
    inp += 'double "double/zero"=0\n'
    inp += 'array double "doublearr/empty"=\n'
    inp += 'array double "doublearr/one"=0.5\n'
    inp += 'array double "doublearr/two"=0.5,-0.25\n'
    inp += 'raw "raw/empty"=\n'
    inp += 'raw "raw/normal"=aGVsbG8=\n'
    inp += 'raw "raw/special"=AAMFCg==\n'
    inp += 'string "string/empty"=""\n'
    inp += 'string "string/normal"="hello"\n'
    inp += 'string "string/special"="\\x00\\x03\\x05\\n"\n'
    inp += 'string "string/quoted"="\\"a\\""\n'
    inp += 'array string "stringarr/empty"=\n'
    inp += 'array string "stringarr/one"="hello"\n'
    inp += 'array string "stringarr/two"="hello","world\\n"\n'

    fp = StringIO(inp)
    assert storage.loadPersistent(fp=fp) is None

    dispatcher._queueOutgoing.assert_has_calls([call(ANY, None, None)] * 23)

    entry_notifier.notifyEntry.assert_has_calls(
        [call(ANY, ANY, ANY, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)] * 23
    )

    assert Value.makeBoolean(True) == storage.getEntryValue("boolean/true")
    assert Value.makeBoolean(False) == storage.getEntryValue("boolean/false")
    assert Value.makeDouble(-1.5) == storage.getEntryValue("double/neg")
    assert Value.makeDouble(0.0) == storage.getEntryValue("double/zero")
    assert Value.makeDouble(1.3e8) == storage.getEntryValue("double/big")
    assert Value.makeString("") == storage.getEntryValue("string/empty")
    assert Value.makeString("hello") == storage.getEntryValue("string/normal")
    assert Value.makeString("\0\3\5\n") == storage.getEntryValue("string/special")
    assert Value.makeString('"a"') == storage.getEntryValue("string/quoted")
    assert Value.makeRaw(b"") == storage.getEntryValue("raw/empty")
    assert Value.makeRaw(b"hello") == storage.getEntryValue("raw/normal")
    assert Value.makeRaw(b"\0\3\5\n") == storage.getEntryValue("raw/special")
    assert Value.makeBooleanArray([]) == storage.getEntryValue("booleanarr/empty")
    assert Value.makeBooleanArray([True]) == storage.getEntryValue("booleanarr/one")
    assert Value.makeBooleanArray([True, False]) == storage.getEntryValue(
        "booleanarr/two"
    )
    assert Value.makeDoubleArray([]) == storage.getEntryValue("doublearr/empty")
    assert Value.makeDoubleArray([0.5]) == storage.getEntryValue("doublearr/one")
    assert Value.makeDoubleArray([0.5, -0.25]) == storage.getEntryValue("doublearr/two")
    assert Value.makeStringArray([]) == storage.getEntryValue("stringarr/empty")
    assert Value.makeStringArray(["hello"]) == storage.getEntryValue("stringarr/one")
    assert Value.makeStringArray(["hello", "world\n"]) == storage.getEntryValue(
        "stringarr/two"
    )
    assert Value.makeBoolean(True) == storage.getEntryValue("\0\3\5\n")
    assert Value.makeBoolean(True) == storage.getEntryValue("CaseSensitive/KeyName")


def test_LoadPersistentWarn(storage_empty, dispatcher, entry_notifier):
    storage = storage_empty

    fp = StringIO('[NetworkTables Storage 3.0]\nboolean "foo"=foo\n')

    # EXPECT_CALL(warn,
    #            Warn(2, llvm.StringRef("unrecognized boolean value, not 'True' or 'False'")))
    assert storage.loadPersistent(fp=fp) is None

    assert len(storage.m_entries) == 0
    assert len(storage.m_idmap) == 0

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_ProcessIncomingEntryAssign0(
    storage_empty, dispatcher, entry_notifier, is_server, conn
):
    storage = storage_empty
    value = Value.makeDouble(1.0)

    entry_id = 0xFFFF if is_server else 0

    storage.processIncoming(Message.entryAssign("foo", entry_id, 0, value, 0), conn)

    if is_server:
        # id assign message reply generated on the server sent to everyone
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryAssign("foo", 0, 0, value, 0), None, None)]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls([call(0, "foo", value, NT_NOTIFY_NEW)])


def test_ProcessIncomingEntryAssign1(
    storage_populate_one, dispatcher, entry_notifier, is_server, conn
):
    storage = storage_populate_one
    value = Value.makeDouble(1.0)

    storage.processIncoming(Message.entryAssign("foo", 0, 1, value, 0), conn)

    # EXPECT_CALL(*conn, proto_rev()).WillRepeatedly(Return(0x0300u))
    if is_server:
        # server broadcasts new value to all *other* connections
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryAssign("foo", 0, 1, value, 0), None, conn)]
        )
    else:
        assert dispatcher._queueOutgoing.call_count == 0

    entry_notifier.notifyEntry.assert_has_calls(
        [call(0, "foo", value, NT_NOTIFY_UPDATE)]
    )


def test_ProcessIncomingEntryAssignIgnore(
    storage_populate_one, dispatcher, entry_notifier, is_server, conn
):
    storage = storage_populate_one
    value = Value.makeDouble(1.0)

    storage.processIncoming(Message.entryAssign("foo", 0xFFFF, 1, value, 0), conn)

    assert dispatcher._queueOutgoing.call_count == 0
    assert entry_notifier.notifyEntry.call_count == 0


def test_ProcessIncomingEntryAssignWithFlags(
    storage_populate_one, dispatcher, entry_notifier, is_server, conn
):
    storage = storage_populate_one
    value = Value.makeDouble(1.0)

    storage.processIncoming(Message.entryAssign("foo", 0, 1, value, 0x2), conn)

    # EXPECT_CALL(*conn, proto_rev()).WillRepeatedly(Return(0x0300u))
    if is_server:
        # server broadcasts new value/flags to all *other* connections
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.entryAssign("foo", 0, 1, value, 0x2), None, conn)]
        )

        entry_notifier.notifyEntry.assert_has_calls(
            [call(0, "foo", value, NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS)]
        )

    else:
        # client forces flags back when an assign message is received for an
        # existing entry with different flags
        dispatcher._queueOutgoing.assert_has_calls(
            [call(Message.flagsUpdate(0, 0), None, None)]
        )

        entry_notifier.notifyEntry.assert_has_calls(
            [call(0, "foo", value, NT_NOTIFY_UPDATE)]
        )


def test_DeleteCheckHandle(storage_populate_one, dispatcher, entry_notifier, is_server):
    storage = storage_populate_one

    handle = storage.getEntryId("foo")
    storage.deleteEntry("foo")
    storage.setEntryTypeValue("foo", Value.makeBoolean(True))

    handle2 = storage.getEntryId("foo")
    assert handle == handle2


def test_DeletedEntryFlags(storage_populate_one, dispatcher, entry_notifier, is_server):
    storage = storage_populate_one

    handle = storage.getEntryId("foo")
    storage.setEntryFlags("foo", 2)
    storage.deleteEntry("foo")

    assert storage.getEntryFlags("foo") == 0
    assert storage.getEntryFlags(handle) == 0
    storage.setEntryFlags("foo", 4)
    storage.setEntryFlags(handle, 4)
    assert storage.getEntryFlags("foo") == 0
    assert storage.getEntryFlags(handle) == 0


def test_DeletedDeleteAllEntries(
    storage_populate_one, dispatcher, entry_notifier, is_server
):
    storage = storage_populate_one

    storage.deleteEntry("foo")
    assert dispatcher._queueOutgoing.call_count == (1 if is_server else 0)

    storage.deleteAllEntries()

    assert dispatcher._queueOutgoing.call_count == (1 if is_server else 0)


def test_DeletedGetEntries(storage_populate_one, dispatcher, entry_notifier, is_server):
    storage = storage_populate_one

    storage.deleteEntry("foo")
    assert storage.getEntries("", 0) == []



# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

#
# These tests are adapted from ntcore's test suite
#

from _pynetworktables._impl.constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
)
from _pynetworktables._impl.value import Value


def test_Boolean():
    v = Value.makeBoolean(False)
    assert NT_BOOLEAN == v.type
    assert not v.value

    v = Value.makeBoolean(True)
    assert NT_BOOLEAN == v.type
    assert v.value


def test_Double():
    v = Value.makeDouble(0.5)
    assert NT_DOUBLE == v.type
    assert 0.5 == v.value

    v = Value.makeDouble(0.25)
    assert NT_DOUBLE == v.type
    assert 0.25 == v.value


def test_String():
    v = Value.makeString("hello")
    assert NT_STRING == v.type
    assert "hello" == v.value

    v = Value.makeString("goodbye")
    assert NT_STRING == v.type
    assert "goodbye" == v.value


def test_Raw():
    v = Value.makeRaw(b"hello")
    assert NT_RAW == v.type
    assert b"hello" == v.value

    v = Value.makeRaw(b"goodbye")
    assert NT_RAW == v.type
    assert b"goodbye" == v.value


def test_BooleanArray():
    v = Value.makeBooleanArray([True, False, True])
    assert NT_BOOLEAN_ARRAY == v.type
    assert (True, False, True) == v.value


def test_DoubleArray():
    v = Value.makeDoubleArray([0.5, 0.25, 0.5])
    assert NT_DOUBLE_ARRAY == v.type
    assert (0.5, 0.25, 0.5) == v.value


def test_StringArray():
    v = Value.makeStringArray(["hello", "goodbye", "string"])
    assert NT_STRING_ARRAY == v.type
    assert ("hello", "goodbye", "string") == v.value


def test_MixedComparison():

    v2 = Value.makeBoolean(True)
    v3 = Value.makeDouble(0.5)
    assert v2 != v3  # boolean vs double


def test_BooleanComparison():
    v1 = Value.makeBoolean(True)
    v2 = Value.makeBoolean(True)
    assert v1 == v2
    v2 = Value.makeBoolean(False)
    assert v1 != v2


def test_DoubleComparison():
    v1 = Value.makeDouble(0.25)
    v2 = Value.makeDouble(0.25)
    assert v1 == v2
    v2 = Value.makeDouble(0.5)
    assert v1 != v2


def test_StringComparison():
    v1 = Value.makeString("hello")
    v2 = Value.makeString("hello")
    assert v1 == v2
    v2 = Value.makeString("world")
    # different contents
    assert v1 != v2
    v2 = Value.makeString("goodbye")
    # different size
    assert v1 != v2


def test_BooleanArrayComparison():
    v1 = Value.makeBooleanArray([1, 0, 1])
    v2 = Value.makeBooleanArray((1, 0, 1))
    assert v1 == v2

    # different contents
    v2 = Value.makeBooleanArray([1, 1, 1])
    assert v1 != v2

    # different size
    v2 = Value.makeBooleanArray([True, False])
    assert v1 != v2


def test_DoubleArrayComparison():
    v1 = Value.makeDoubleArray([0.5, 0.25, 0.5])
    v2 = Value.makeDoubleArray((0.5, 0.25, 0.5))
    assert v1 == v2

    # different contents
    v2 = Value.makeDoubleArray([0.5, 0.5, 0.5])
    assert v1 != v2

    # different size
    v2 = Value.makeDoubleArray([0.5, 0.25])
    assert v1 != v2


def test_StringArrayComparison():
    v1 = Value.makeStringArray(["hello", "goodbye", "string"])
    v2 = Value.makeStringArray(("hello", "goodbye", "string"))
    assert v1 == v2

    # different contents
    v2 = Value.makeStringArray(["hello", "goodby2", "string"])
    assert v1 != v2

    # different sized contents
    v2 = Value.makeStringArray(["hello", "goodbye2", "string"])
    assert v1 != v2

    # different size
    v2 = Value.makeStringArray(["hello", "goodbye"])
    assert v1 != v2


#
# Additional Python tests
#


def test_unicode():
    # copyright symbol
    v1 = Value.makeString(u"\xA9")
    assert v1.value == u"\xA9"


def test_bytearray():
    v1 = Value.makeRaw(bytearray(b"\x01\x02\x00"))
    assert v1.type == NT_RAW
    assert v1.value == bytearray(b"\x01\x02\x00")
    assert v1.value == b"\x01\x02\x00"



#
# Useful fixtures
#

from contextlib import contextmanager
from threading import Condition

log_datefmt = "%H:%M:%S"
log_format = "%(asctime)s:%(msecs)03d %(levelname)-8s: %(name)-8s: %(message)s"

import logging

logging.basicConfig(level=logging.DEBUG, format=log_format, datefmt=log_datefmt)


logger = logging.getLogger("conftest")

import pytest

from _pynetworktables import NetworkTables, NetworkTablesInstance

#
# Fixtures for a usable in-memory version of networktables
#


@pytest.fixture(scope="function", params=[True, False])
def verbose_logging(request):
    return request.param


@pytest.fixture(scope="function", params=[True, False])
def nt(request):
    """Starts/stops global networktables instance for testing"""
    NetworkTables.startTestMode(server=request.param)

    yield NetworkTables

    NetworkTables.shutdown()


@pytest.fixture(scope="function")
def entry_notifier(nt):
    return nt._api.entry_notifier


@pytest.fixture(scope="function")
def conn_notifier(nt):
    return nt._api.conn_notifier


@pytest.fixture(scope="function")
def nt_flush(nt):
    """Flushes NT key notifications"""

    def _flush():
        assert nt._api.waitForEntryListenerQueue(1.0)
        assert nt._api.waitForConnectionListenerQueue(1.0)

    return _flush


#
# Live NT instance fixtures
#


class NtTestBase(NetworkTablesInstance):
    """
    Object for managing a live pair of NT server/client
    """

    _wait_lock = None
    _testing_verbose_logging = True

    def shutdown(self):
        logger.info("shutting down %s", self.__class__.__name__)
        NetworkTablesInstance.shutdown(self)
        if self._wait_lock is not None:
            self._wait_init_listener()

    def disconnect(self):
        self._api.dispatcher.stop()

    def _init_common(self, proto_rev):
        # This resets the instance to be independent
        self.shutdown()
        self._api.dispatcher.setDefaultProtoRev(proto_rev)
        self.proto_rev = proto_rev

        if self._testing_verbose_logging:
            self.enableVerboseLogging()
        # self._wait_init()

    def _init_server(self, proto_rev, server_port=0):
        self._init_common(proto_rev)

        self.port = server_port

    def _init_client(self, proto_rev):
        self._init_common(proto_rev)

    def _wait_init(self):
        self._wait_lock = Condition()
        self._wait = 0
        self._wait_init_listener()

    def _wait_init_listener(self):
        self._api.addEntryListener(
            "",
            self._wait_cb,
            NetworkTablesInstance.NotifyFlags.NEW
            | NetworkTablesInstance.NotifyFlags.UPDATE
            | NetworkTablesInstance.NotifyFlags.DELETE
            | NetworkTablesInstance.NotifyFlags.FLAGS,
        )

    def _wait_cb(self, *args):
        with self._wait_lock:
            self._wait += 1
            # logger.info('Wait callback, got: %s', args)
            self._wait_lock.notify()

    @contextmanager
    def expect_changes(self, count):
        """Use this on the *other* instance that you're making
        changes on, to wait for the changes to propagate to the
        other instance"""

        if self._wait_lock is None:
            self._wait_init()

        with self._wait_lock:
            self._wait = 0

        logger.info("Begin actions")
        yield
        logger.info("Waiting for %s changes", count)

        with self._wait_lock:
            result, msg = (
                self._wait_lock.wait_for(lambda: self._wait == count, 4),
                "Timeout waiting for %s changes (got %s)" % (count, self._wait),
            )
            logger.info("expect_changes: %s %s", result, msg)
            assert result, msg


# Each test should cover each NT version combination
# 0x0200 -> 0x0300
# 0x0300 -> 0x0200
# 0x0300 -> 0x0300


@pytest.fixture(params=[0x0200, 0x0300])
def nt_server(request, verbose_logging):
    class NtServer(NtTestBase):

        _test_saved_port = None
        _testing_verbose_logging = verbose_logging

        def start_test(self):
            logger.info("NtServer::start_test")

            # Restore server port on restart
            if self._test_saved_port is not None:
                self.port = self._test_saved_port
                self._api.dispatcher.setDefaultProtoRev(request.param)

            if verbose_logging:
                self.enableVerboseLogging()

            self.startServer(listenAddress="127.0.0.1", port=self.port)

            assert self._api.dispatcher.m_server_acceptor.waitForStart(timeout=1)
            self.port = self._api.dispatcher.m_server_acceptor.m_port
            self._test_saved_port = self.port

    server = NtServer()
    server._init_server(request.param)
    yield server
    server.shutdown()


@pytest.fixture(params=[0x0200, 0x0300])
def nt_client(request, nt_server, verbose_logging):
    class NtClient(NtTestBase):

        _testing_verbose_logging = verbose_logging

        def start_test(self):
            if verbose_logging:
                self.enableVerboseLogging()
            self.setNetworkIdentity("C1")
            self._api.dispatcher.setDefaultProtoRev(request.param)
            self.startClient(("127.0.0.1", nt_server.port))

    client = NtClient()
    client._init_client(request.param)
    yield client
    client.shutdown()


@pytest.fixture(params=[0x0300])  # don't bother with other proto versions
def nt_client2(request, nt_server, verbose_logging):
    class NtClient(NtTestBase):

        _testing_verbose_logging = verbose_logging

        def start_test(self):
            if verbose_logging:
                self.enableVerboseLogging()
            self._api.dispatcher.setDefaultProtoRev(request.param)
            self.setNetworkIdentity("C2")
            self.startClient(("127.0.0.1", nt_server.port))

    client = NtClient()
    client._init_client(request.param)
    yield client
    client.shutdown()


@pytest.fixture
def nt_live(nt_server, nt_client):
    """This fixture automatically starts the client and server"""

    nt_server.start_test()
    nt_client.start_test()

    return nt_server, nt_client



# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

#
# These tests are adapted from ntcore's test suite
#

import threading


def test_Polled(nt_server, nt_client):
    nt_server_api = nt_server._api

    if nt_server.proto_rev == 0x0200 or nt_client.proto_rev == 0x0200:
        # this is annoying to test because of the reconnect
        return

    # set up the poller
    poller = nt_server_api.createConnectionListenerPoller()
    handle = nt_server_api.addPolledConnectionListener(poller, False)

    # trigger a connect event
    nt_server.start_test()
    nt_client.start_test()

    # get the event
    assert nt_server_api.waitForConnectionListenerQueue(1.0)

    result, timed_out = nt_server_api.pollConnectionListener(poller, 1.0)
    assert not timed_out
    assert len(result) == 1
    assert handle == result[0][0]
    assert result[0][1].connected
    del result[:]

    # trigger a disconnect event
    nt_client.shutdown()

    # get the event
    assert nt_server_api.waitForConnectionListenerQueue(1.0)

    result, timed_out = nt_server_api.pollConnectionListener(poller, 0.1)
    assert not timed_out
    assert len(result) == 1
    assert handle == result[0][0]
    assert not result[0][1].connected


def test_Threaded(nt_server, nt_client):
    nt_server_api = nt_server._api

    if nt_server.proto_rev == 0x0200 or nt_client.proto_rev == 0x0200:
        # this is annoying to test because of the reconnect
        return

    result_cond = threading.Condition()
    result = []

    def _server_cb(event):
        with result_cond:
            result.append(event)
            result_cond.notify()

    nt_server_api.addConnectionListener(_server_cb, False)

    # trigger a connect event
    nt_server.start_test()
    nt_client.start_test()

    with result_cond:
        result_cond.wait(0.5)

    # get the event
    assert len(result) == 1
    # assert handle == result[0].listener
    assert result[0].connected
    del result[:]

    # trigger a disconnect event
    nt_client.shutdown()

    with result_cond:
        result_cond.wait(0.5)

    # get the event
    assert len(result) == 1
    # assert handle == result[0].listener
    assert not result[0].connected



# validated: 2018-01-06 DV 2287281066f6 cpp/Message.cpp cpp/Message.h

from collections import namedtuple

from .constants import (
    kKeepAlive,
    kClientHello,
    kProtoUnsup,
    kServerHello,
    kServerHelloDone,
    kClientHelloDone,
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    kExecuteRpc,
    kRpcResponse,
    kClearAllMagic,
    NT_VTYPE2RAW,
    NT_RAW2VTYPE,
)


class Message(
    namedtuple("Message", ["type", "str", "value", "id", "flags", "seq_num_uid"])
):
    __slots__ = ()

    _empty_msgtypes = (kKeepAlive, kServerHelloDone, kClientHelloDone)

    @classmethod
    def keepAlive(cls):
        return cls(kKeepAlive, None, None, None, None, None)

    @classmethod
    def clientHello(cls, proto_rev, identity):
        return cls(kClientHello, identity, None, proto_rev, None, None)

    @classmethod
    def protoUnsup(cls, proto_rev):
        return cls(kProtoUnsup, None, None, proto_rev, None, None)

    @classmethod
    def serverHelloDone(cls):
        return cls(kServerHelloDone, None, None, None, None, None)

    @classmethod
    def serverHello(cls, flags, identity):
        return cls(kServerHello, identity, None, None, flags, None)

    @classmethod
    def clientHelloDone(cls):
        return cls(kClientHelloDone, None, None, None, None, None)

    @classmethod
    def entryAssign(cls, name, msg_id, seq_num_uid, value, flags):
        return cls(kEntryAssign, name, value, msg_id, flags, seq_num_uid)

    @classmethod
    def entryUpdate(cls, entry_id, seq_num_uid, value):
        return cls(kEntryUpdate, None, value, entry_id, None, seq_num_uid)

    @classmethod
    def flagsUpdate(cls, msg_id, flags):
        return cls(kFlagsUpdate, None, None, msg_id, flags, None)

    @classmethod
    def entryDelete(cls, entry_id):
        return cls(kEntryDelete, None, None, entry_id, None, None)

    @classmethod
    def clearEntries(cls):
        return cls(kClearEntries, None, None, kClearAllMagic, None, None)

    @classmethod
    def executeRpc(cls, rpc_id, call_uid, params):
        return cls(kExecuteRpc, params, None, rpc_id, None, call_uid)

    @classmethod
    def rpcResponse(cls, rpc_id, call_uid, result):
        return cls(kRpcResponse, result, None, rpc_id, None, call_uid)

    @classmethod
    def read(cls, rstream, codec, get_entry_type) -> "Message":
        msgtype = rstream.read(1)

        msg_str = None
        value = None
        msg_id = None
        flags = None
        seq_num_uid = None

        # switch type
        if msgtype in cls._empty_msgtypes:
            pass

        # python optimization: entry updates tend to occur more than
        # anything else, so check this first
        elif msgtype == kEntryUpdate:
            if codec.proto_rev >= 0x0300:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryUpdate)
                value_type = NT_RAW2VTYPE.get(rstream.read(1))
            else:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryUpdate)
                value_type = get_entry_type(msg_id)

            value = codec.read_value(value_type, rstream)

        elif msgtype == kClientHello:
            (msg_id,) = rstream.readStruct(codec.clientHello)
            if msg_id >= 0x0300:
                msg_str = codec.read_string_v3(rstream)

        elif msgtype == kProtoUnsup:
            (msg_id,) = rstream.readStruct(codec.protoUnsup)

        elif msgtype == kServerHello:
            (flags,) = rstream.readStruct(codec.serverHello)
            msg_str = codec.read_string(rstream)

        elif msgtype == kEntryAssign:
            msg_str = codec.read_string(rstream)
            value_type = NT_RAW2VTYPE.get(rstream.read(1))
            if codec.proto_rev >= 0x0300:
                msg_id, seq_num_uid, flags = rstream.readStruct(codec.entryAssign)
            else:
                msg_id, seq_num_uid = rstream.readStruct(codec.entryAssign)
                flags = 0

            value = codec.read_value(value_type, rstream)

        elif msgtype == kFlagsUpdate:
            msg_id, flags = rstream.readStruct(codec.flagsUpdate)

        elif msgtype == kEntryDelete:
            (msg_id,) = rstream.readStruct(codec.entryDelete)

        elif msgtype == kClearEntries:
            (msg_id,) = rstream.readStruct(codec.clearEntries)
            if msg_id != kClearAllMagic:
                raise ValueError("Bad magic")

        elif msgtype == kExecuteRpc:
            msg_id, seq_num_uid = rstream.readStruct(codec.executeRpc)
            msg_str = codec.read_string(rstream)

        elif msgtype == kRpcResponse:
            msg_id, seq_num_uid = rstream.readStruct(codec.rpcResponse)
            msg_str = codec.read_string(rstream)

        else:
            raise ValueError("Unrecognized message type %s" % msgtype)

        return cls(msgtype, msg_str, value, msg_id, flags, seq_num_uid)

    def write(self, out, codec):
        msgtype = self.type

        # switch type
        if msgtype in self._empty_msgtypes:
            out.append(msgtype)

        elif msgtype == kClientHello:
            proto_rev = self.id
            out += (msgtype, codec.clientHello.pack(proto_rev))

            if proto_rev >= 0x0300:
                codec.write_string_v3(self.str, out)

        elif msgtype == kProtoUnsup:
            out += (msgtype, codec.protoUnsup.pack(self.id))

        elif msgtype == kServerHello:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.serverHello.pack(self.flags))
                codec.write_string(self.str, out)

        elif msgtype == kEntryAssign:
            out.append(msgtype)
            codec.write_string(self.str, out)

            value = self.value
            if codec.proto_rev >= 0x0300:
                sb = codec.entryAssign.pack(self.id, self.seq_num_uid, self.flags)
            else:
                sb = codec.entryAssign.pack(self.id, self.seq_num_uid)
            out += (NT_VTYPE2RAW[value.type], sb)

            codec.write_value(value, out)

        elif msgtype == kEntryUpdate:
            value = self.value
            if codec.proto_rev >= 0x0300:
                out += (
                    msgtype,
                    codec.entryUpdate.pack(self.id, self.seq_num_uid),
                    NT_VTYPE2RAW[value.type],
                )
            else:
                out += (msgtype, codec.entryUpdate.pack(self.id, self.seq_num_uid))

            codec.write_value(value, out)

        elif msgtype == kFlagsUpdate:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.flagsUpdate.pack(self.id, self.flags))

        elif msgtype == kEntryDelete:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.entryDelete.pack(self.id))

        elif msgtype == kClearEntries:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.clearEntries.pack(self.id))

        elif msgtype == kExecuteRpc:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.executeRpc.pack(self.id, self.seq_num_uid))
                codec.write_string(self.str, out)

        elif msgtype == kRpcResponse:
            if codec.proto_rev >= 0x0300:
                out += (msgtype, codec.rpcResponse.pack(self.id, self.seq_num_uid))
                codec.write_string(self.str, out)

        else:
            raise ValueError("Internal error: bad value type %s" % self.type)



import pytest
from _pynetworktables.entry import NetworkTableEntry


@pytest.mark.parametrize(
    "data, expected_result",
    [
        ("a", True),
        (1, True),
        (1.0, True),
        (1.0j, False),
        (False, True),
        (b"1", True),
        (bytearray(), True),
        ((), ValueError),
        ([], ValueError),
    ],
)
def test_isValidType(data, expected_result):
    if isinstance(expected_result, bool):
        assert NetworkTableEntry.isValidDataType(data) == expected_result
    else:
        with pytest.raises(expected_result):
            NetworkTableEntry.isValidDataType(data)



#
# These tests stand up a separate client and server instance of
# networktables and tests the 'real' user API to ensure that it
# works correctly
#

from __future__ import print_function

import pytest

import logging

logger = logging.getLogger("test")


# test defaults
def doc(nt):
    t = nt.getTable("nope")

    assert t.getBoolean("b", None) is None
    assert t.getNumber("n", None) is None
    assert t.getString("s", None) is None
    assert t.getBooleanArray("ba", None) is None
    assert t.getNumberArray("na", None) is None
    assert t.getStringArray("sa", None) is None
    assert t.getValue("v", None) is None

    assert t.getBoolean("b", True) is True
    assert t.getNumber("n", 1) == 1
    assert t.getString("s", "sss") == "sss"
    assert t.getBooleanArray("ba", (True,)) == (True,)
    assert t.getNumberArray("na", (1,)) == (1,)
    assert t.getStringArray("sa", ("ss",)) == ("ss",)
    assert t.getValue("v", "vvv") == "vvv"


def do(nt1, nt2, t):

    t1 = nt1.getTable(t)
    with nt2.expect_changes(8):
        t1.putBoolean("bool", True)
        t1.putNumber("number1", 1)
        t1.putNumber("number2", 1.5)
        t1.putString("string", "string")
        t1.putString("unicode", u"\xA9")  # copyright symbol
        t1.putBooleanArray("ba", (True, False))
        t1.putNumberArray("na", (1, 2))
        t1.putStringArray("sa", ("s", "t"))

    t2 = nt2.getTable(t)
    assert t2.getBoolean("bool", None) is True
    assert t2.getNumber("number1", None) == 1
    assert t2.getNumber("number2", None) == 1.5
    assert t2.getString("string", None) == "string"
    assert t2.getString("unicode", None) == u"\xA9"  # copyright symbol
    assert t2.getBooleanArray("ba", None) == (True, False)
    assert t2.getNumberArray("na", None) == (1, 2)
    assert t2.getStringArray("sa", None) == ("s", "t")

    # Value testing
    with nt2.expect_changes(6):
        t1.putValue("v_b", False)
        t1.putValue("v_n1", 2)
        t1.putValue("v_n2", 2.5)
        t1.putValue("v_s", "ssss")
        t1.putValue("v_s2", u"\xA9")

        t1.putValue("v_v", 0)

    print(t2.getKeys())
    assert t2.getBoolean("v_b", None) is False
    assert t2.getNumber("v_n1", None) == 2
    assert t2.getNumber("v_n2", None) == 2.5
    assert t2.getString("v_s", None) == "ssss"
    assert t2.getString("v_s2", None) == u"\xA9"
    assert t2.getValue("v_v", None) == 0

    # Ensure that updating values work!
    with nt2.expect_changes(8):
        t1.putBoolean("bool", False)
        t1.putNumber("number1", 2)
        t1.putNumber("number2", 2.5)
        t1.putString("string", "sss")
        t1.putString("unicode", u"\u2122")  # (tm)
        t1.putBooleanArray("ba", (False, True, False))
        t1.putNumberArray("na", (2, 1))
        t1.putStringArray("sa", ("t", "s"))

    t2 = nt2.getTable(t)
    assert t2.getBoolean("bool", None) is False
    assert t2.getNumber("number1", None) == 2
    assert t2.getNumber("number2", None) == 2.5
    assert t2.getString("string", None) == "sss"
    assert t2.getString("unicode", None) == u"\u2122"
    assert t2.getBooleanArray("ba", None) == (False, True, False)
    assert t2.getNumberArray("na", None) == (2, 1)
    assert t2.getStringArray("sa", None) == ("t", "s")

    # Try out deletes -- but NT2 doesn't support them
    if nt2.proto_rev == 0x0300:
        if nt1.proto_rev == 0x0300:
            with nt2.expect_changes(1):
                t1.delete("bool")

            assert t2.getBoolean("bool", None) == None
        else:
            t1.delete("bool")

            with nt2.expect_changes(1):
                t1.putBoolean("ooo", True)

            assert t2.getBoolean("bool", None) is False

    else:
        t1.delete("bool")

        with nt2.expect_changes(1):
            t1.putBoolean("ooo", True)

        assert t2.getBoolean("bool", None) is False


def test_basic(nt_live):

    nt_server, nt_client = nt_live

    assert nt_server.isServer()
    assert not nt_client.isServer()

    doc(nt_client)
    doc(nt_server)

    # server -> client
    do(nt_server, nt_client, "server2client")

    # client -> server
    do(nt_client, nt_server, "client2server")

    assert nt_client.isConnected()
    assert nt_server.isConnected()


def test_reconnect(nt_live):

    nt_server, nt_client = nt_live

    with nt_server.expect_changes(1):
        ct = nt_client.getTable("t")
        ct.putBoolean("foo", True)

    st = nt_server.getTable("t")
    assert st.getBoolean("foo", None) == True

    # Client disconnect testing
    nt_client.shutdown()

    logger.info("Shutdown the client")

    with nt_client.expect_changes(1):
        nt_client.start_test()
        ct = nt_client.getTable("t")

    assert ct.getBoolean("foo", None) == True

    # Server disconnect testing
    nt_server.shutdown()
    logger.info("Shutdown the server")

    # synchronization change: if the client doesn't touch the entry locally,
    # then it won't get transferred back to the server on reconnect. Touch
    # it here to ensure that it comes back
    ct.putBoolean("foo", True)

    with nt_server.expect_changes(1):
        nt_server.start_test()

    st = nt_server.getTable("t")
    assert st.getBoolean("foo", None) == True



from __future__ import print_function
from io import BytesIO

import pytest

from _pynetworktables._impl.support.leb128 import (
    size_uleb128,
    encode_uleb128,
    read_uleb128,
)


def test_size():

    # Testing Plan:
    # (1) 128 ^ n ............ need (n+1) bytes
    # (2) 128 ^ n * 64 ....... need (n+1) bytes
    # (3) 128 ^ (n+1) - 1 .... need (n+1) bytes

    assert 1 == size_uleb128(0)  # special case

    assert 1 == size_uleb128(0x1)
    assert 1 == size_uleb128(0x40)
    assert 1 == size_uleb128(0x7F)

    assert 2 == size_uleb128(0x80)
    assert 2 == size_uleb128(0x2000)
    assert 2 == size_uleb128(0x3FFF)

    assert 3 == size_uleb128(0x4000)
    assert 3 == size_uleb128(0x100000)
    assert 3 == size_uleb128(0x1FFFFF)

    assert 4 == size_uleb128(0x200000)
    assert 4 == size_uleb128(0x8000000)
    assert 4 == size_uleb128(0xFFFFFFF)

    assert 5 == size_uleb128(0x10000000)
    assert 5 == size_uleb128(0x40000000)
    assert 5 == size_uleb128(0x7FFFFFFF)


def test_wikipedia_example():

    i = 624485
    result_bytes = bytes(bytearray([0xE5, 0x8E, 0x26]))

    bio = BytesIO(encode_uleb128(i))
    print(result_bytes)
    assert bio.read() == result_bytes
    bio.seek(0)

    result_i = read_uleb128(bio)

    assert result_i == i


@pytest.mark.parametrize(
    "i", [0, 1, 16, 42, 65, 0xFF, 0xFFF, 0xFFFFFFFF, 0x123456789, 100000000000000000000]
)
def test_roundtrip(i):

    bio = BytesIO(encode_uleb128(i))
    print(bio.read())
    bio.seek(0)

    r = read_uleb128(bio)

    print(r)

    assert r == i



# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

#
# These tests are adapted from ntcore's test suite
#

import pytest

from _pynetworktables._impl.constants import (
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_NEW,
    NT_NOTIFY_DELETE,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS,
)

from _pynetworktables._impl.entry_notifier import EntryNotifier
from _pynetworktables._impl.value import Value


@pytest.fixture
def notifier():
    n = EntryNotifier(verbose=True)
    n.start()
    yield n
    n.stop()


def generateNotifications(notifier):
    # All flags combos that can be generated by Storage
    flags = [
        # "normal" notifications
        NT_NOTIFY_NEW,
        NT_NOTIFY_DELETE,
        NT_NOTIFY_UPDATE,
        NT_NOTIFY_FLAGS,
        NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS,
        # immediate notifications are always "new"
        NT_NOTIFY_IMMEDIATE | NT_NOTIFY_NEW,
        # local notifications can be of any flag combo
        NT_NOTIFY_LOCAL | NT_NOTIFY_NEW,
        NT_NOTIFY_LOCAL | NT_NOTIFY_DELETE,
        NT_NOTIFY_LOCAL | NT_NOTIFY_UPDATE,
        NT_NOTIFY_LOCAL | NT_NOTIFY_FLAGS,
        NT_NOTIFY_LOCAL | NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS,
    ]

    # Generate across keys
    keys = ["/foo/bar", "/baz", "/boo"]

    val = Value.makeDouble(1)

    # Provide unique key indexes for each key
    keyindex = 5
    for key in keys:
        for flag in flags:
            notifier.notifyEntry(keyindex, key, val, flag)

        keyindex += 1


def test_PollEntryMultiple(notifier):
    poller1 = notifier.createPoller()
    poller2 = notifier.createPoller()
    poller3 = notifier.createPoller()
    h1 = notifier.addPolledById(poller1, 6, NT_NOTIFY_NEW)
    h2 = notifier.addPolledById(poller2, 6, NT_NOTIFY_NEW)
    h3 = notifier.addPolledById(poller3, 6, NT_NOTIFY_UPDATE)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results1, timed_out = notifier.poll(poller1, 0)
    assert not timed_out
    results2, timed_out = notifier.poll(poller2, 0)
    assert not timed_out
    results3, timed_out = notifier.poll(poller3, 0)
    assert not timed_out

    assert len(results1) == 2
    for h, result in results1:
        print(result)
        assert h == h1

    assert len(results2) == 2
    for h, result in results2:
        print(result)
        assert h == h2

    assert len(results3) == 2
    for h, result in results3:
        print(result)
        assert h == h3


def test_PollEntryBasic(notifier):
    poller = notifier.createPoller()
    g1 = notifier.addPolledById(poller, 6, NT_NOTIFY_NEW)
    g2 = notifier.addPolledById(poller, 6, NT_NOTIFY_DELETE)
    g3 = notifier.addPolledById(poller, 6, NT_NOTIFY_UPDATE)
    g4 = notifier.addPolledById(poller, 6, NT_NOTIFY_FLAGS)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    timed_out = False
    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out

    g1count = 0
    g2count = 0
    g3count = 0
    g4count = 0
    for h, result in results:
        print(h, result)
        assert result.name == "/baz"
        assert result.value == Value.makeDouble(1)
        assert result.local_id == 6

        if h == g1:
            g1count += 1
            assert (result.flags & NT_NOTIFY_NEW) != 0
        elif h == g2:
            g2count += 1
            assert (result.flags & NT_NOTIFY_DELETE) != 0
        elif h == g3:
            g3count += 1
            assert (result.flags & NT_NOTIFY_UPDATE) != 0
        elif h == g4:
            g4count += 1
            assert (result.flags & NT_NOTIFY_FLAGS) != 0
        else:
            assert False, "unknown listener index"

    assert g1count == 2
    assert g2count == 1  # NT_NOTIFY_DELETE
    assert g3count == 2
    assert g4count == 2


def test_PollEntryImmediate(notifier):
    poller = notifier.createPoller()
    notifier.addPolledById(poller, 6, NT_NOTIFY_NEW | NT_NOTIFY_IMMEDIATE)
    notifier.addPolledById(poller, 6, NT_NOTIFY_NEW)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out
    print(results)
    assert len(results) == 4


def test_PollEntryLocal(notifier):
    poller = notifier.createPoller()
    notifier.addPolledById(poller, 6, NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
    notifier.addPolledById(poller, 6, NT_NOTIFY_NEW)

    assert notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out
    print(results)
    assert len(results) == 6


def test_PollPrefixMultiple(notifier):
    poller1 = notifier.createPoller()
    poller2 = notifier.createPoller()
    poller3 = notifier.createPoller()
    h1 = notifier.addPolled(poller1, "/foo", NT_NOTIFY_NEW)
    h2 = notifier.addPolled(poller2, "/foo", NT_NOTIFY_NEW)
    h3 = notifier.addPolled(poller3, "/foo", NT_NOTIFY_UPDATE)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results1, timed_out = notifier.poll(poller1, 0)
    assert not timed_out
    results2, timed_out = notifier.poll(poller2, 0)
    assert not timed_out
    results3, timed_out = notifier.poll(poller3, 0)
    assert not timed_out

    assert len(results1) == 2
    for h, result in results1:
        print(result)
        assert h == h1

    assert len(results2) == 2
    for h, result in results2:
        print(result)
        assert h == h2

    assert len(results3) == 2
    for h, result in results3:
        print(result)
        assert h == h3


def test_PollPrefixBasic(notifier):
    poller = notifier.createPoller()
    g1 = notifier.addPolled(poller, "/foo", NT_NOTIFY_NEW)
    g2 = notifier.addPolled(poller, "/foo", NT_NOTIFY_DELETE)
    g3 = notifier.addPolled(poller, "/foo", NT_NOTIFY_UPDATE)
    g4 = notifier.addPolled(poller, "/foo", NT_NOTIFY_FLAGS)
    notifier.addPolled(poller, "/bar", NT_NOTIFY_NEW)
    notifier.addPolled(poller, "/bar", NT_NOTIFY_DELETE)
    notifier.addPolled(poller, "/bar", NT_NOTIFY_UPDATE)
    notifier.addPolled(poller, "/bar", NT_NOTIFY_FLAGS)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out

    g1count = 0
    g2count = 0
    g3count = 0
    g4count = 0
    for h, result in results:
        print(result)
        assert result.name.startswith("/foo")
        assert result.value == Value.makeDouble(1)

        if h == g1:
            g1count += 1
            assert (result.flags & NT_NOTIFY_NEW) != 0
        elif h == g2:
            g2count += 1
            assert (result.flags & NT_NOTIFY_DELETE) != 0
        elif h == g3:
            g3count += 1
            assert (result.flags & NT_NOTIFY_UPDATE) != 0
        elif h == g4:
            g4count += 1
            assert (result.flags & NT_NOTIFY_FLAGS) != 0
        else:
            assert False, "unknown listener index"

    assert g1count == 2
    assert g2count == 1  # NT_NOTIFY_DELETE
    assert g3count == 2
    assert g4count == 2


def test_PollPrefixImmediate(notifier):
    poller = notifier.createPoller()
    notifier.addPolled(poller, "/foo", NT_NOTIFY_NEW | NT_NOTIFY_IMMEDIATE)
    notifier.addPolled(poller, "/foo", NT_NOTIFY_NEW)

    assert not notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out
    print(results)
    assert len(results) == 4


def test_PollPrefixLocal(notifier):
    poller = notifier.createPoller()
    notifier.addPolled(poller, "/foo", NT_NOTIFY_NEW | NT_NOTIFY_LOCAL)
    notifier.addPolled(poller, "/foo", NT_NOTIFY_NEW)

    assert notifier.m_local_notifiers

    generateNotifications(notifier)

    assert notifier.waitForQueue(1.0)

    results, timed_out = notifier.poll(poller, 0)
    assert not timed_out
    print(results)
    assert len(results) == 6



#!/usr/bin/env python3
#
# This example uses matplotlib to graph data received from the logger example.
# It is assumed that the key being logged is a number array, and the first
# item of that array is the Timer.getFPGATimestamp() on the robot.
#
# One example way to send the data on the robot is via something like
# wpilib.SmartDashboard.putNumberArray([time, data1, data2, ...])
#

import json
import sys

import matplotlib.pyplot as plt
import numpy as np

if __name__ == "__main__":
    with open(sys.argv[1]) as fp:
        data = json.load(fp)

    if len(data) == 0:
        print("No data received")
        exit(0)

    # Change the time to
    offset = data[0][0]
    for d in data:
        d[0] -= offset

    print("Received", len(data), "rows of data, total time was %.3f seconds" % d[0])

    # Transform the data into a numpy array to make it easier to use
    data = np.array(data)

    # This allows you to use data[N] to refer to each column of data individually
    data = data.transpose()

    # This silly plot graphs
    # - x: data[0]: this is time
    # - y: data[1]: column1
    #
    # - x2: data[0]: Time yet again
    # - y2: data[1] - data[2]: subtracts columns from each other (something numpy allows)
    plt.plot(data[0], data[1], data[0], data[1] - data[2])
    plt.title("Encoder error")

    plt.show()



from networktables.util import ntproperty, ChooserControl
import pytest


def test_autoupdatevalue(nt):

    # tricksy: make sure that this works *before* initialization
    # of network tables happens!
    nt.shutdown()

    foo = nt.getGlobalAutoUpdateValue("/SmartDashboard/foo", True, True)
    assert foo.value == True
    assert foo.get() == True

    nt.startTestMode()

    assert foo.value == True
    assert foo.get() == True

    t = nt.getTable("/SmartDashboard")
    assert t.getBoolean("foo", None) == True
    t.putBoolean("foo", False)

    assert foo.value == False


def test_ntproperty(nt, nt_flush):
    class Foo(object):
        robotTime = ntproperty(
            "/SmartDashboard/robotTime", 0, writeDefault=False, inst=nt
        )
        dsTime = ntproperty("/SmartDashboard/dsTime", 0, writeDefault=True, inst=nt)
        testArray = ntproperty(
            "/SmartDashboard/testArray", [1, 2, 3], writeDefault=True, inst=nt
        )

    f = Foo()

    t = nt.getTable("/SmartDashboard")

    assert f.robotTime == 0
    assert t.getNumber("robotTime", None) == 0

    f.robotTime = 2
    assert t.getNumber("robotTime", None) == 2

    t.putNumber("robotTime", 4)
    assert f.robotTime == 4

    assert f.testArray == (1, 2, 3)
    f.testArray = [4, 5, 6]
    assert f.testArray == (4, 5, 6)


def test_ntproperty_emptyarray(nt):
    with pytest.raises(TypeError):

        class Foo1(object):
            testArray = ntproperty(
                "/SmartDashboard/testArray", [], writeDefault=True, inst=nt
            )

    with pytest.raises(TypeError):

        class Foo2(object):
            testArray = ntproperty(
                "/SmartDashboard/testArray", [], writeDefault=False, inst=nt
            )


def test_ntproperty_multitest(nt):
    """
    Checks to see that ntproperties still work between NT restarts
    """

    class Foo(object):
        robotTime = ntproperty(
            "/SmartDashboard/robotTime", 0, writeDefault=False, inst=nt
        )
        dsTime = ntproperty("/SmartDashboard/dsTime", 0, writeDefault=True, inst=nt)

    for i in range(3):
        print("Iteration", i)

        f = Foo()

        t = nt.getTable("/SmartDashboard")

        assert f.robotTime == 0
        assert f.dsTime == 0

        assert t.getNumber("robotTime", None) == 0
        assert t.getNumber("dsTime", None) == 0

        f.robotTime = 2
        assert t.getNumber("robotTime", None) == 2
        assert t.getNumber("dsTime", None) == 0

        t.putNumber("robotTime", 4)
        assert f.robotTime == 4
        assert f.dsTime == 0

        nt.shutdown()
        nt.startTestMode()


def test_chooser_control(nt):

    c = ChooserControl("Autonomous Mode", inst=nt)

    assert c.getChoices() == ()
    assert c.getSelected() is None

    c.setSelected("foo")
    assert c.getSelected() == "foo"

    t = nt.getTable("/SmartDashboard/Autonomous Mode")
    assert t.getString("selected", None) == "foo"

    t.putStringArray("options", ("option1", "option2"))
    assert c.getChoices() == ("option1", "option2")



#
# These tests stand up a separate client and server instance of
# networktables and tests the 'real' user API to ensure that it
# works correctly
#

from __future__ import print_function

import logging

logger = logging.getLogger("test")


#
# Distinction between the following test cases:
# -> instance.shutdown clears the storage for the nt instance
# -> instance.disconnect only shuts down the network connection, storage is retained
#

#
# Writes before connection
#


def test_sync_pre_client_writes_value(nt_server, nt_client):
    """
    Client writes value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | client foo=1     |
    | connected        | foo=1
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    ct.putString("foo", "1")

    with nt_server.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_pre_server_writes_value(nt_server, nt_client):
    """
    Server writes value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | server foo=1     |
    | connected        | foo=1
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    st.putString("foo", "1")

    with nt_client.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_pre_both_write_values(nt_server, nt_client):
    """
    Both write values

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | server foo=1     |
    | client foo=2     |
    | connected        | foo=2           | client wins
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    st.putString("foo", "1")
    ct.putString("foo", "2")

    with nt_server.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"


def test_sync_srestart_client_writes(nt_server, nt_client):
    """
    Server restart; Client wrote value locally

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | client foo=1     |
    | connected        | foo=1
    | server restart   |
    | connected        | foo=1
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    ct.putString("foo", "1")

    with nt_server.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_server.shutdown()

    st = nt_server.getTable("table")
    assert st.getString("foo", None) == None

    with nt_server.expect_changes(1):
        nt_server.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_srestart_server_writes(nt_server, nt_client):
    """
    Server restart; Server wrote value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | server foo=1     |
    | connected        | foo=1
    | server restart   |
    | connected        | foo deleted     | Not an intuitive result
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    st.putString("foo", "1")

    with nt_client.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_server.shutdown()

    st = nt_server.getTable("table")
    assert st.getString("foo", None) == None

    # Required otherwise we don't have anything to wait for...
    ct.putNumber("ignored", 1)

    with nt_server.expect_changes(1):
        nt_server.start_test()

    assert ct.getString("foo", None) == None
    assert st.getString("foo", None) == None


def test_sync_srestart_both_writes(nt_server, nt_client):
    """
    Server restart; Both write value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | server foo=2     | foo=2
    | server restart   |
    | connected        | foo=2
    """

    nt_server.start_test()
    nt_client.start_test()

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    with nt_client.expect_changes(1):
        st.putString("foo", "2")

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"

    nt_server.shutdown()

    st = nt_server.getTable("table")
    assert st.getString("foo", None) == None

    with nt_server.expect_changes(1):
        nt_server.start_test()

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"


def test_sync_crestart_client_writes(nt_server, nt_client):
    """
    Client restart

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | client foo=1     |
    | connected        | foo=1
    | client restart   |
    | connected        | foo=1
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    ct.putString("foo", "1")

    with nt_server.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.shutdown()
    ct = nt_client.getTable("table")

    assert ct.getString("foo", None) == None

    with nt_client.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_crestart_server_writes(nt_server, nt_client):
    """
    Client restart

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | server foo=1     |
    | connected        | foo=1
    | client restart   |
    | connected        | foo=1
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    st.putString("foo", "1")

    with nt_client.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.shutdown()
    ct = nt_client.getTable("table")

    assert ct.getString("foo", None) == None

    with nt_client.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_crestart_server_writes_late(nt_server, nt_client):
    """
    Client restart

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | server foo=1     |
    | connected        | foo=1
    | client restart   |
    | server foo=2     |
    | connected        | foo=2
    """

    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    st.putString("foo", "1")

    with nt_client.expect_changes(1):
        nt_server.start_test()
        nt_client.start_test()

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.shutdown()
    ct = nt_client.getTable("table")

    assert ct.getString("foo", None) == None
    st.putString("foo", "2")

    with nt_client.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"


def test_sync_disconnect_write_by_server(nt_server, nt_client):
    """
    Server update during disconnect - server initiated value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | server foo=1     | foo=1
    | disconnect
    | server foo=2
    | connected        | foo=2
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_client.expect_changes(1):
        st.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.disconnect()
    st.putString("foo", "2")

    with nt_client.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"


def test_sync_disconnect_write_by_client(nt_server, nt_client):
    """
    Server update during disconnect - client initiated value

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | disconnect
    | server foo=2
    | connected        | foo=1           | client wins
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.disconnect()
    st.putString("foo", "2")

    with nt_server.expect_changes(1):
        nt_client.start_test()

    # client wins
    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"


def test_sync_disconnect_write_by_client2(nt_server, nt_client):
    """
    Client update during disconnect

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | disconnect
    | client foo=2
    | connected        | foo=2
    | client foo=3     | foo=3
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.disconnect()
    ct.putString("foo", "2")

    with nt_server.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"

    with nt_server.expect_changes(1):
        ct.putString("foo", "3")

    # more writes succeed
    assert ct.getString("foo", None) == "3"
    assert st.getString("foo", None) == "3"


def test_sync_disconnect_write_by_both(nt_server, nt_client):
    """
    Both update during disconnect

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | disconnect
    | client foo=2
    | server foo=3
    | connected        | foo=2
    | client foo=4     | foo=4
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    nt_client.disconnect()
    ct.putString("foo", "2")
    st.putString("foo", "3")

    with nt_server.expect_changes(1):
        nt_client.start_test()

    # client wins
    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"

    with nt_server.expect_changes(1):
        ct.putString("foo", "4")

    # more writes succeed
    assert ct.getString("foo", None) == "4"
    assert st.getString("foo", None) == "4"


def test_sync_disconnect_write_by_both_prev(nt_server, nt_client):
    """
    Client and server updates during disconnect (both previously written)

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | server foo=2     | foo=2
    | disconnect
    | client foo=3
    | connected        | foo=3           | issue #270?
    | client foo=4     | foo=4
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    with nt_client.expect_changes(1):
        st.putString("foo", "2")

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"

    nt_client.disconnect()
    ct.putString("foo", "3")

    with nt_server.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "3"
    assert st.getString("foo", None) == "3"

    with nt_server.expect_changes(1):
        ct.putString("foo", "4")

    # more writes succeed
    assert ct.getString("foo", None) == "4"
    assert st.getString("foo", None) == "4"


def test_sync_disconnect_write_by_both_both_prev(nt_server, nt_client):
    """
    Client and server updates during disconnect (both previously written)

    | Action           | Global NT state | Notes
    | ---------------- | --------------- | -----
    | connected
    | client foo=1     | foo=1
    | server foo=2     | foo=2
    | disconnect
    | server foo=3
    | client foo=4
    | connected        | foo=4           | issue #270?
    | client foo=5     | foo=5
    """
    ct = nt_client.getTable("table")
    st = nt_server.getTable("table")

    nt_server.start_test()
    nt_client.start_test()

    with nt_server.expect_changes(1):
        ct.putString("foo", "1")

    assert ct.getString("foo", None) == "1"
    assert st.getString("foo", None) == "1"

    with nt_client.expect_changes(1):
        st.putString("foo", "2")

    assert ct.getString("foo", None) == "2"
    assert st.getString("foo", None) == "2"

    nt_client.disconnect()
    st.putString("foo", "3")
    ct.putString("foo", "4")

    with nt_server.expect_changes(1):
        nt_client.start_test()

    assert ct.getString("foo", None) == "4"
    assert st.getString("foo", None) == "4"

    with nt_server.expect_changes(1):
        ct.putString("foo", "5")

    # more writes succeed
    assert ct.getString("foo", None) == "5"
    assert st.getString("foo", None) == "5"



#
# These tests are adapted from ntcore's test suite
#

from __future__ import print_function

from io import BytesIO

import pytest

from _pynetworktables._impl.message import Message
from _pynetworktables._impl.value import Value
from _pynetworktables._impl.tcpsockets.tcp_stream import TCPStream, StreamEOF
from _pynetworktables._impl.wire import WireCodec


class ReadStream(TCPStream):
    def __init__(self, fp):
        self.m_rdsock = fp


@pytest.fixture(params=[0x0200, 0x0300])
def proto_rev(request):
    return request.param


@pytest.fixture
def v_round_trip(proto_rev):

    codec = WireCodec(proto_rev)

    def _fn(v, minver=0x0200):
        if codec.proto_rev < minver:
            return

        out = []
        fp = BytesIO()
        rstream = ReadStream(fp)

        codec.write_value(v, out)
        fp.write(b"".join(out))
        fp.seek(0)

        vv = codec.read_value(v.type, rstream)

        with pytest.raises(StreamEOF):
            rstream.read(1)

        assert v == vv

    return _fn


# for each value type, test roundtrip


def test_wire_boolean(v_round_trip):
    v_round_trip(Value.makeBoolean(True))


def test_wire_double(v_round_trip):
    v_round_trip(Value.makeDouble(0.5))


def test_wire_string1(v_round_trip):
    v_round_trip(Value.makeString(""))


def test_wire_string2(v_round_trip):
    v_round_trip(Value.makeString("Hi there"))


def test_wire_raw1(v_round_trip):
    v_round_trip(Value.makeRaw(b""), minver=0x0300)


def test_wire_raw2(v_round_trip):
    v_round_trip(Value.makeRaw(b"\x00\xff\x78"), minver=0x0300)


def test_wire_boolArray1(v_round_trip):
    v_round_trip(Value.makeBooleanArray([]))


def test_wire_boolArray2(v_round_trip):
    v_round_trip(Value.makeBooleanArray([True, False]))


def test_wire_boolArray3(v_round_trip):
    v_round_trip(Value.makeBooleanArray([True] * 255))


def test_wire_doubleArray1(v_round_trip):
    v_round_trip(Value.makeDoubleArray([]))


def test_wire_doubleArray2(v_round_trip):
    v_round_trip(Value.makeDoubleArray([0, 1]))


def test_wire_doubleArray3(v_round_trip):
    v_round_trip(Value.makeDoubleArray([0] * 255))


def test_wire_stringArray1(v_round_trip):
    v_round_trip(Value.makeStringArray([]))


def test_wire_stringArray2(v_round_trip):
    v_round_trip(Value.makeStringArray(["hi", "there"]))


def test_wire_stringArray3(v_round_trip):
    v_round_trip(Value.makeStringArray(["hi"] * 255))


def test_wire_rpc1(v_round_trip):
    v_round_trip(Value.makeRpc(""), minver=0x0300)


def test_wire_rpc2(v_round_trip):
    v_round_trip(Value.makeRpc("Hi there"), minver=0x0300)


# Try out the various message types
@pytest.fixture
def msg_round_trip(proto_rev):

    codec = WireCodec(proto_rev)

    def _fn(msg: Message, minver=0x0200, exclude=None):

        out = []
        fp = BytesIO()
        rstream = ReadStream(fp)

        if codec.proto_rev < minver:
            # The codec won't have the correct struct set if
            # the version isn't supported
            msg.write(out, codec)
            assert not out
            return

        msg.write(out, codec)
        fp.write(b"".join(out))
        fp.seek(0)

        mm = Message.read(rstream, codec, lambda x: msg.value.type)

        with pytest.raises(StreamEOF):
            rstream.read(1)

        # In v2, some fields aren't copied over, so we exclude them
        # by overwriting those indices and recreating the read message
        if exclude:
            args = list(mm)
            for e in exclude:
                args[e] = msg[e]
            mm = Message(*args)

        assert msg == mm

    return _fn


def test_wire_keepAlive(msg_round_trip):
    msg_round_trip(Message.keepAlive())


def test_wire_clientHello(msg_round_trip):
    msg_round_trip(Message.clientHello(0x0200, "Hi"), exclude=[1])


def test_wire_clientHelloV3(msg_round_trip):
    msg_round_trip(Message.clientHello(0x0300, "Hi"))


def test_wire_protoUnsup(msg_round_trip):
    msg_round_trip(Message.protoUnsup(0x0300))


def test_wire_serverHelloDone(msg_round_trip):
    msg_round_trip(Message.serverHelloDone())


def test_wire_serverHello(msg_round_trip):
    msg_round_trip(Message.serverHello(0x01, "Hi"), minver=0x0300)


def test_wire_clientHelloDone(msg_round_trip):
    msg_round_trip(Message.clientHelloDone())


def test_wire_entryAssign1(msg_round_trip, proto_rev):
    exclude = [] if proto_rev >= 0x0300 else [4]
    value = Value.makeBoolean(True)
    msg_round_trip(
        Message.entryAssign("Hi", 0x1234, 0x4321, value, 0x00), exclude=exclude
    )


def test_wire_entryAssign2(msg_round_trip, proto_rev):
    exclude = [] if proto_rev >= 0x0300 else [4]
    value = Value.makeString("Oh noes")
    msg_round_trip(
        Message.entryAssign("Hi", 0x1234, 0x4321, value, 0x00), exclude=exclude
    )


def test_wire_entryUpdate1(msg_round_trip, proto_rev):
    exclude = [] if proto_rev >= 0x0300 else [4]
    value = Value.makeBoolean(True)
    msg_round_trip(Message.entryUpdate(0x1234, 0x4321, value), exclude=exclude)


def test_wire_entryUpdate2(msg_round_trip, proto_rev):
    exclude = [] if proto_rev >= 0x0300 else [4]
    value = Value.makeString("Oh noes")
    msg_round_trip(Message.entryUpdate(0x1234, 0x4321, value), exclude=exclude)


def test_wire_flagsUpdate(msg_round_trip):
    msg_round_trip(Message.flagsUpdate(0x1234, 0x42), minver=0x0300)


def test_wire_entryDelete(msg_round_trip):
    msg_round_trip(Message.entryDelete(0x1234), minver=0x0300)


def test_wire_clearEntries(msg_round_trip):
    msg_round_trip(Message.clearEntries(), minver=0x0300)


def test_wire_executeRpc(msg_round_trip):
    msg_round_trip(Message.executeRpc(0x1234, 0x4321, "parameter"), minver=0x0300)


def test_wire_rpcResponse(msg_round_trip):
    msg_round_trip(Message.rpcResponse(0x1234, 0x4321, "parameter"), minver=0x0300)


# Various invalid unicode
def test_decode_invalid_string(proto_rev):
    codec = WireCodec(proto_rev)

    if proto_rev == 0x0200:
        prefix = b"\x00\x1a"
    else:
        prefix = b"\x1a"

    fp = BytesIO(prefix + b"\x00\xa7>\x03eWithJoystickCommandV2")
    rstream = ReadStream(fp)

    s = "INVALID UTF-8: b'\\x00\\xa7>\\x03eWithJoystickCommandV2'"

    assert codec.read_string(rstream) == s



#!/usr/bin/env python3
#
# A simplified fuzzer that can be used to try and discover errors in the
# NetworkTables implementation.
#
# DON'T USE THIS ON A ROBOT CONNECTED TO REAL THINGS! Not all NetworkTables
# implementations are robust, and may crash or malfunction when subjected
# to this fuzzer
#

#
# TODO: Do more evil things to tease out deadlock/other bugs
#

ip = "127.0.0.1"
port = 1735

import errno
import threading
import socket
import os
from random import choice, randint

import time


def random_bytes(n):
    return bytes(os.urandom(n))


def sendbytes(s, a):
    s.send(bytes(a))


# valid message types
message_types = [0x00, 0x01, 0x02, 0x03, 0x10, 0x11]

num_threads = 16


def fuzz_any():
    ret = [choice(message_types)]

    ret += os.urandom(128)
    return ret


def fuzz_singlebyte():
    ret = [choice(message_types)]
    return ret


def fuzz_assign():

    ret = [0x10]

    # string:
    # 2 bytes len
    # n bytes content

    l = randint(0, 255)
    ret += [0, l]

    ret += os.urandom(l)

    # byte of type id
    ret += [randint(0, 20)]

    # two bytes, entry id
    ret += [0, randint(0, 255)]

    # two bytes, sequence
    ret += [randint(0, 255), randint(0, 255)]

    # some value
    ret += [0, 0]

    return ret


def fuzz_update():

    ret = [0x11]

    # entry id
    ret += [0x00, randint(0, 10)]

    # data type
    ret += [randint(0, 5)]

    # data
    ret += os.urandom(128)
    return ret


def fuzz_gibberish():
    return os.urandom(128)


def fuzz_dumb():
    return [
        randint(0, 5),
        randint(0, 5),
        randint(0, 5),
        randint(0, 5),
        randint(0, 5),
        randint(0, 5),
        randint(0, 5),
    ]


fuzz_routines = [fuzz_assign, fuzz_singlebyte, fuzz_update, fuzz_gibberish, fuzz_dumb]


def fuzz_thread():

    i = 0

    while True:

        print("Iteration", i)
        i += 1

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(1)

        try:
            print("Opening socket")
            s.connect((ip, port))

            sendbytes(s, choice(fuzz_routines)())
            s.recv(1)

        except socket.timeout:
            print("Socket timed out, try again")

        except socket.error as e:
            if e.errno != errno.ECONNRESET:
                raise
        finally:
            print("Closing socket")
            s.close()


threads = [threading.Thread(target=fuzz_thread) for i in range(0, num_threads)]

for t in threads:
    t.start()

    time.sleep(0.2)



from _pynetworktables._impl.storage import _Entry


def test_sequence_numbers():

    e = _Entry("name", 0, None)

    #
    # Test rollover
    #

    e.seq_num = 0xFFFE
    e.increment_seqnum()
    assert e.seq_num == 0xFFFF

    e.increment_seqnum()
    assert e.seq_num == 0

    #
    # test Entry.isSeqNewerThan
    # -> operator >
    #

    e.seq_num = 10
    assert e.isSeqNewerThan(20) == False

    e.seq_num = 20
    assert e.isSeqNewerThan(10) == True

    e.seq_num = 50000
    assert e.isSeqNewerThan(10) == False

    e.seq_num = 10
    assert e.isSeqNewerThan(50000) == True

    e.seq_num = 20
    assert e.isSeqNewerThan(20) == False

    e.seq_num = 50000
    assert e.isSeqNewerThan(50000) == False

    #
    # test Entry.isSeqNewerOrEqual
    # -> operator >=
    #

    e.seq_num = 10
    assert e.isSeqNewerOrEqual(20) == False

    e.seq_num = 20
    assert e.isSeqNewerOrEqual(10) == True

    e.seq_num = 50000
    assert e.isSeqNewerOrEqual(10) == False

    e.seq_num = 10
    assert e.isSeqNewerOrEqual(50000) == True

    e.seq_num = 20
    assert e.isSeqNewerOrEqual(20) == True

    e.seq_num = 50000
    assert e.isSeqNewerOrEqual(50000) == True



#!/usr/bin/env python3
#
# This is a simple working example of how to use pynetworktables to log data
# to a JSON file. The file is stored in the current working directory, and
# the filename is the current time combined with the event name and match number
#
# The data really can be anything, but this logs a single key. The way it's
# intended to be used is gather your data together, stick it in an array
# along with the current timestamp, then send it all over networktables
# as a number array.
#
# One example way to send the data on the robot is via something like
# wpilib.SmartDashboard.putNumberArray([time, data1, data2, ...])
#
# The reason this example uses that method is to ensure that all of the data
# received is transmitted at the same time. If you used multiple keys to send
# this data instead, the data could be slightly out of sync with each other.
#

from networktables import NetworkTables
from networktables.util import ntproperty

import json
import os.path
import queue
import sys
import time
import threading

import logging

logger = logging.getLogger("logger")

# FMSControlData bitfields
ENABLED_FIELD = 1 << 0
AUTO_FIELD = 1 << 1
TEST_FIELD = 1 << 2
EMERGENCY_STOP_FIELD = 1 << 3
FMS_ATTACHED_FIELD = 1 << 4
DS_ATTACHED_FIELD = 1 << 5


def translate_control_word(value):
    value = int(value)
    if value & ENABLED_FIELD == 0:
        return "disabled"
    if value & AUTO_FIELD:
        return "auto"
    if value & TEST_FIELD:
        return "test"
    else:
        return "teleop"


class DataLogger:

    # Change this key to whatever NT key you want to log
    log_key = "/SmartDashboard/log_data"

    # Data file where robot IP is stored so you don't have to keep typing it
    cache_file = ".robot"

    matchNumber = ntproperty("/FMSInfo/MatchNumber", 0, False)
    eventName = ntproperty("/FMSInfo/EventName", "unknown", False)

    def __init__(self):
        self.queue = queue.Queue()
        self.mode = "disabled"
        self.data = []
        self.lock = threading.Lock()

    def connectionListener(self, connected, info):
        # set our robot to 'disabled' if the connection drops so that we can
        # guarantee the data gets written to disk
        if not connected:
            self.valueChanged("/FMSInfo/FMSControlData", 0, False)

    def valueChanged(self, key, value, isNew):

        if key == "/FMSInfo/FMSControlData":

            mode = translate_control_word(value)

            with self.lock:
                last = self.mode
                self.mode = mode

                data = self.data
                self.data = []

            logger.info("Robot mode: %s -> %s", last, mode)

            # This example only stores on auto -> disabled transition. Change it
            # to whatever it is that you need for logging
            if last == "auto":

                tm = time.strftime("%Y%m%d-%H%M-%S")
                name = "%s-%s-%s.json" % (tm, self.eventName, int(self.matchNumber))
                logger.info("New file: %s (%d items received)", name, len(data))

                # We don't write the file from within the NetworkTables callback,
                # because we don't want to block the thread. Instead, write it
                # to a queue along with the filename so it can be written
                # from somewhere else
                self.queue.put((name, data))

        elif key == self.log_key:
            if self.mode != "disabled":
                with self.lock:
                    self.data.append(value)

    def run(self):

        # Determine what IP to connect to
        try:
            server = sys.argv[1]

            # Save the robot ip
            if not os.path.exists(self.cache_file):
                with open(self.cache_file, "w") as fp:
                    fp.write(server)
        except IndexError:
            try:
                with open(self.cache_file) as fp:
                    server = fp.read().strip()
            except IOError:
                print("Usage: logger.py [10.xx.yy.2]")
                return

        logger.info("NT server set to %s", server)
        NetworkTables.initialize(server=server)

        # Use listeners to receive the data
        NetworkTables.addConnectionListener(
            self.connectionListener, immediateNotify=True
        )
        NetworkTables.addEntryListener(self.valueChanged)

        # When new data is queued, write it to disk
        while True:
            name, data = self.queue.get()
            with open(name, "w") as fp:
                json.dump(data, fp)


if __name__ == "__main__":

    log_datefmt = "%H:%M:%S"
    log_format = "%(asctime)s:%(msecs)03d %(levelname)-8s: %(name)-20s: %(message)s"

    logging.basicConfig(level=logging.DEBUG, datefmt=log_datefmt, format=log_format)

    dl = DataLogger()
    dl.run()



#
# Synchronization analysis scenarios
#

import pytest


@pytest.fixture
def scenario(nt_server, nt_client, nt_client2):
    st = nt_server.getTable("table")
    ct1 = nt_client.getTable("table")
    ct2 = nt_client2.getTable("table")

    nt_server.start_test()

    st.putString("ServerOnly", "0")

    ct1.putString("Client1Only", "1")
    ct1.putString("SC1Shared", "1")
    ct1.putString("ClientShared", "1")

    ct2.putString("Client2Only", "2")
    ct2.putString("SC2Shared", "2")

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(4):  # receives own updates
            nt_client.start_test()

    with nt_server.expect_changes(2):
        with nt_client2.expect_changes(6):  # receives own updates
            with nt_client.expect_changes(2):
                nt_client2.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=0,
        Client1Only=1,
        SC1Shared=1,
        ClientShared=1,
        Client2Only=2,
        SC2Shared=2,
    )

    return nt_server, nt_client, nt_client2, st, ct1, ct2


def check_results(st, ct1, ct2, **kwargs):
    for k, v in kwargs.items():
        if v is not None:
            v = str(v)
        assert st.getString(k, None) == v
        assert ct1.getString(k, None) == v
        assert ct2.getString(k, None) == v


def test_scenario_1(scenario):
    """
    Single client network drop (disconnect + reconnect)

    Client1 disconnects from Server
    Client1 updates Client1Only=11, SC1Shared=11, ClientShared=11
    Server updates ServerOnly=10, SC1Shared=10, SC2Shared=10
    Client2 updates Client2Only=12, SC2Shared=12, ClientShared=12
    Client1 reconnects to server
    """
    nt_server, nt_client, nt_client2, st, ct1, ct2 = scenario

    nt_client.disconnect()

    ct1.putString("Client1Only", "11")
    ct1.putString("SC1Shared", "11")
    ct1.putString("ClientShared", "11")

    with nt_client2.expect_changes(3):
        st.putString("ServerOnly", "10")
        st.putString("SC1Shared", "10")
        st.putString("SC2Shared", "10")

    with nt_server.expect_changes(3):
        ct2.putString("Client2Only", "12")
        ct2.putString("SC2Shared", "12")
        ct2.putString("ClientShared", "12")

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            with nt_client2.expect_changes(3):
                nt_client.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=10,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=11,
        SC2Shared=12,
        ClientShared=11,
    )


def test_scenario_2(scenario):
    """
    Server network drop (both clients disconnect + reconnect)

    Client1 disconnects from Server
    Client2 disconnects from Server
    Client1 updates Client1Only=11, SC1Shared=11, ClientShared=11
    Client2 updates Client2Only=12, SC2Shared=12, ClientShared=12
    Server updates ServerOnly=10, SC1Shared=10, SC2Shared=10
    Client1 reconnects to server
    Client2 reconnects to server
    """
    nt_server, nt_client, nt_client2, st, ct1, ct2 = scenario

    nt_client.disconnect()
    nt_client2.disconnect()

    ct1.putString("Client1Only", "11")
    ct1.putString("SC1Shared", "11")
    ct1.putString("ClientShared", "11")

    ct2.putString("Client2Only", "12")
    ct2.putString("SC2Shared", "12")
    ct2.putString("ClientShared", "12")

    st.putString("ServerOnly", "10")
    st.putString("SC1Shared", "10")
    st.putString("SC2Shared", "10")

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            nt_client.start_test()

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            with nt_client2.expect_changes(3):
                nt_client2.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=10,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=11,
        SC2Shared=12,
        ClientShared=12,
    )


def test_scenario_3(scenario):
    """
    Server reboot, reconnect before server-local code makes changes

    Server restarts (disconnecting both Client1 and Client2)
    Client1 updates Client1Only=11, SC1Shared=11, ClientShared=11
    Client2 updates Client2Only=12, SC2Shared=12, ClientShared=12
    Client1 reconnects to Server
    Client2 reconnects to Server
    Server updates ServerOnly=10, SC1Shared=10, SC2Shared=10
    """
    nt_server, nt_client, nt_client2, st, ct1, ct2 = scenario

    nt_server.shutdown()
    st = nt_server.getTable("table")

    ct1.putString("Client1Only", "11")
    ct1.putString("SC1Shared", "11")
    ct1.putString("ClientShared", "11")

    ct2.putString("Client2Only", "12")
    ct2.putString("SC2Shared", "12")
    ct2.putString("ClientShared", "12")

    # needed to assure reconnection order
    nt_client2.disconnect()

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(6):
            nt_server.start_test()

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            with nt_client2.expect_changes(5):
                nt_client2.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=None,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=11,
        SC2Shared=12,
        ClientShared=12,
    )

    with nt_client.expect_changes(3):
        with nt_client2.expect_changes(3):
            st.putString("ServerOnly", "10")
            st.putString("SC1Shared", "10")
            st.putString("SC2Shared", "10")

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=10,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=10,
        SC2Shared=10,
        ClientShared=12,
    )


def test_scenario_4(scenario):
    """
    Server reboot, reconnect after server-local code makes changes

    Server restarts (disconnecting both Client1 and Client2)
    Server updates ServerOnly=10, SC1Shared=10, SC2Shared=10
    Client1 updates Client1Only=11, SC1Shared=11, ClientShared=11
    Client2 updates Client2Only=12, SC2Shared=12, ClientShared=12
    Client1 reconnects to Server
    Client2 reconnects to Server
    """
    nt_server, nt_client, nt_client2, st, ct1, ct2 = scenario

    nt_server.shutdown()
    st = nt_server.getTable("table")

    st.putString("ServerOnly", "10")
    st.putString("SC1Shared", "10")
    st.putString("SC2Shared", "10")

    ct1.putString("Client1Only", "11")
    ct1.putString("SC1Shared", "11")
    ct1.putString("ClientShared", "11")

    ct2.putString("Client2Only", "12")
    ct2.putString("SC2Shared", "12")
    ct2.putString("ClientShared", "12")

    # needed to assure reconnection order
    nt_client2.disconnect()

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(5):
            nt_server.start_test()

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            with nt_client2.expect_changes(4):
                nt_client2.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=10,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=11,
        SC2Shared=12,
        ClientShared=12,
    )


def test_scenario_5(scenario):
    """
    Single client reboot

    Client1 restarts (disconnecting from Server)
    Client1 updates Client1Only=11, SC1Shared=11, ClientShared=11
    Server updates ServerOnly=10, SC1Shared=10, SC2Shared=10
    Client2 updates Client2Only=12, SC2Shared=12, ClientShared=12
    Client1 reconnects to Server
    """
    nt_server, nt_client, nt_client2, st, ct1, ct2 = scenario

    nt_client.shutdown()
    ct1 = nt_client.getTable("table")

    ct1.putString("Client1Only", "11")
    ct1.putString("SC1Shared", "11")
    ct1.putString("ClientShared", "11")

    with nt_client2.expect_changes(3):
        st.putString("ServerOnly", "10")
        st.putString("SC1Shared", "10")
        st.putString("SC2Shared", "10")

    with nt_server.expect_changes(3):
        ct2.putString("Client2Only", "12")
        ct2.putString("SC2Shared", "12")
        ct2.putString("ClientShared", "12")

    with nt_server.expect_changes(3):
        with nt_client.expect_changes(3):
            with nt_client2.expect_changes(3):
                nt_client.start_test()

    check_results(
        st,
        ct1,
        ct2,
        ServerOnly=10,
        Client1Only=11,
        Client2Only=12,
        SC1Shared=11,
        SC2Shared=12,
        ClientShared=11,
    )



#
# These tests are leftover from the original pynetworktables tests
#

from unittest.mock import call, Mock
import pytest

from _pynetworktables import NetworkTables


@pytest.fixture(scope="function")
def table1(nt):
    return nt.getTable("/test1")


@pytest.fixture(scope="function")
def table2(nt):
    return nt.getTable("/test2")


@pytest.fixture(scope="function")
def table3(nt):
    return nt.getTable("/test3")


@pytest.fixture(scope="function")
def subtable1(nt):
    return nt.getTable("/test2/sub1")


@pytest.fixture(scope="function")
def subtable2(nt):
    return nt.getTable("/test2/sub2")


@pytest.fixture(scope="function")
def subtable3(nt):
    return nt.getTable("/test3/suba")


@pytest.fixture(scope="function")
def subtable4(nt):
    return nt.getTable("/test3/suba/subb")


def test_key_listener_immediate_notify(table1, nt_flush):

    listener1 = Mock()

    table1.putBoolean("MyKey1", True)
    table1.putBoolean("MyKey1", False)
    table1.putBoolean("MyKey2", True)
    table1.putBoolean("MyKey4", False)
    nt_flush()

    table1.addEntryListener(listener1.valueChanged, True, localNotify=True)

    nt_flush()
    listener1.valueChanged.assert_has_calls(
        [
            call(table1, "MyKey1", False, True),
            call(table1, "MyKey2", True, True),
            call(table1, "MyKey4", False, True),
        ],
        True,
    )
    assert len(listener1.mock_calls) == 3
    listener1.reset_mock()

    table1.putBoolean("MyKey", False)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey", False, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey1", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey1", True, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey1", False)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey1", False, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey4", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey4", True, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()


def test_key_listener_not_immediate_notify(table1, nt_flush):

    listener1 = Mock()

    table1.putBoolean("MyKey1", True)
    table1.putBoolean("MyKey1", False)
    table1.putBoolean("MyKey2", True)
    table1.putBoolean("MyKey4", False)

    table1.addEntryListener(listener1.valueChanged, False, localNotify=True)
    assert len(listener1.mock_calls) == 0
    listener1.reset_mock()

    table1.putBoolean("MyKey", False)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey", False, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey1", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey1", True, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey1", False)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey1", False, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey4", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey4", True, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()


def test_specific_key_listener(table1, nt_flush):

    listener1 = Mock()

    table1.addEntryListener(
        listener1.valueChanged, False, key="MyKey1", localNotify=True
    )
    nt_flush()
    assert len(listener1.mock_calls) == 0

    table1.putBoolean("MyKey1", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(table1, "MyKey1", True, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey2", True)
    nt_flush()
    assert len(listener1.mock_calls) == 0


def test_specific_entry_listener(table1, nt_flush):

    listener1 = Mock()
    NotifyFlags = NetworkTables.NotifyFlags

    entry = table1.getEntry("MyKey1")
    entry.addListener(
        listener1.valueChanged, NotifyFlags.NEW | NotifyFlags.UPDATE | NotifyFlags.LOCAL
    )
    nt_flush()
    assert len(listener1.mock_calls) == 0

    table1.putBoolean("MyKey1", True)
    nt_flush()
    listener1.valueChanged.assert_called_once_with(entry, "/test1/MyKey1", True, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    table1.putBoolean("MyKey2", True)
    nt_flush()
    assert len(listener1.mock_calls) == 0


def test_subtable_listener(table2, subtable1, subtable2, nt_flush):

    listener1 = Mock()

    table2.putBoolean("MyKey1", True)
    table2.putBoolean("MyKey1", False)
    table2.addSubTableListener(listener1.valueChanged, localNotify=True)
    table2.putBoolean("MyKey2", True)
    table2.putBoolean("MyKey4", False)

    subtable1.putBoolean("MyKey1", False)

    nt_flush()
    listener1.valueChanged.assert_called_once_with(table2, "sub1", subtable1, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    subtable1.putBoolean("MyKey2", True)
    subtable1.putBoolean("MyKey1", True)
    subtable2.putBoolean("MyKey1", False)

    nt_flush()
    listener1.valueChanged.assert_called_once_with(table2, "sub2", subtable2, True)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()


def test_subsubtable_listener(table3, subtable3, subtable4, nt_flush):
    listener1 = Mock()

    table3.addSubTableListener(listener1.valueChanged, localNotify=True)
    subtable3.addSubTableListener(listener1.valueChanged, localNotify=True)
    subtable4.addEntryListener(listener1.valueChanged, True, localNotify=True)

    subtable4.putBoolean("MyKey1", False)

    nt_flush()
    listener1.valueChanged.assert_has_calls(
        [
            call(table3, "suba", subtable3, True),
            call(subtable3, "subb", subtable4, True),
            call(subtable4, "MyKey1", False, True),
        ],
        True,
    )
    assert len(listener1.mock_calls) == 3
    listener1.reset_mock()

    subtable4.putBoolean("MyKey1", True)

    nt_flush()
    listener1.valueChanged.assert_called_once_with(subtable4, "MyKey1", True, False)
    assert len(listener1.mock_calls) == 1
    listener1.reset_mock()

    listener2 = Mock()

    table3.addSubTableListener(listener2.valueChanged, localNotify=True)
    subtable3.addSubTableListener(listener2.valueChanged, localNotify=True)
    subtable4.addEntryListener(listener2.valueChanged, True, localNotify=True)

    nt_flush()
    listener2.valueChanged.assert_has_calls(
        [
            call(table3, "suba", subtable3, True),
            call(subtable3, "subb", subtable4, True),
            call(subtable4, "MyKey1", True, True),
        ],
        True,
    )
    assert len(listener1.mock_calls) == 0
    assert len(listener2.mock_calls) == 3
    listener2.reset_mock()


def test_global_listener(nt, nt_flush, table1, subtable3):
    listener = Mock()

    nt.addEntryListener(listener)
    listener.assert_not_called()

    table1.putString("t1", "hi")
    subtable3.putString("tt", "y0")
    nt_flush()
    listener.assert_has_calls(
        [call("/test1/t1", "hi", True), call("/test3/suba/tt", "y0", True)]
    )
    listener.reset_mock()

    nt.removeEntryListener(listener)

    table1.putString("s", "1")
    nt_flush()
    listener.assert_not_called()
